import sys
from rhodium import *
from rhodium.ffi import NativeModel

# Provide the DLL/SO file and the function name.
model = NativeModel("lake.so" if sys.platform == "linux" else "lake.dll", "lake_problem")

# List the inputs.  The order matters!
model.parameters = [Parameter("pollution_limit", type="double*"),
                    Parameter("b", default_value=0.42, type="double"),
                    Parameter("q", default_value=2, type="double"),
                    Parameter("mean", default_value=0.02, type="double"),
                    Parameter("stdev", default_value=0.001, type="double"),
                    Parameter("delta", default_value=0.98, type="double")]

# List all outputs.  We use asarg=True to handle the outputs as arguments to the C
# function.
model.responses = [Response("max_P", Response.MINIMIZE, type="double", asarg=True),
                   Response("utility", Response.MAXIMIZE, type="double", asarg=True),
                   Response("inertia", Response.MAXIMIZE, type="double", asarg=True),
                   Response("reliability", Response.MAXIMIZE, type="double", asarg=True)]

# Specify the levers
model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

# Optimize the model using Rhodium
output = optimize(model, "NSGAII", 10000)
print(output)



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import os
import mplcursors
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import numpy as np
import pandas as pd
import seaborn as sns
from collections.abc import Sequence
from scipy.interpolate import griddata
from matplotlib.legend_handler import HandlerPatch
from .config import RhodiumConfig
from .model import Response
from .brush import BrushSet, apply_brush, color_brush, brush_color_map, color_indices

# When set, override the plt.show() method to save.  Intended for CI
# or headless mode.
_figure_output_envvar = "RHODIUM_FIGURE_OUTPUT"

if os.getenv(_figure_output_envvar):
    def show_override(*args, **kwargs):
        directory = os.getenv(_figure_output_envvar)
        
        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)
        
        filename = os.path.join(directory, f"figure_{plt.gcf().number}.png")
        plt.savefig(filename)
        print(f"{_figure_output_envvar} set, saving figure to {filename}")
        
    plt.show = show_override

def _combine_keys(*args):
    result = []
    result_set = set()

    for arg in args:
        if arg is None:
            continue

        if hasattr(arg, "__iter__") and not isinstance(arg, str):
            for key in arg:
                if key not in result_set:
                    result.append(key)
                    result_set.add(key)
        elif arg not in result_set:
            result.append(arg)
            result_set.add(arg)

    return result

def _is_color(c, expected_size=None):
    if isinstance(c, str):
        return True
    
    if not isinstance(c, Sequence):
        return False
    
    if expected_size:
        if len(c) != expected_size:
            raise ValueError(f"expected color array to have {expected_size} elements, given {len(c)}")
        
    for color in c:
        if not mpl.colors.is_color_like(color):
            return False
        
    return True
   
class HandlerSizeLegend(HandlerPatch):
    def __call__(self, legend, orig_handle, fontsize, handlebox):
        pass

    def create_artists(self, legend, orig_handle,
                       xdescent, ydescent, width, height, fontsize, trans):
        p1 = mpatches.Circle(xy=(0.2 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent),
                             radius=(height*0.25)/2)
        self.update_prop(p1, orig_handle, legend)
        p1.set_transform(trans)

        p2 = mpatches.Circle(xy=(0.66 * width - 0.5 * xdescent, 0.5 * height - 0.5 * ydescent),
                             radius=(height*1.0)/2)
        self.update_prop(p2, orig_handle, legend)
        p2.set_transform(trans)

        return [p1, p2]

def scatter3d(model, data,
              x=None,
              y=None,
              z=None,
              c=None,
              s=None,
              s_range=(10, 50),
              show_colorbar=True,
              show_legend=False,
              interactive=False,
              brush=None,
              pick_handler=None,
              **kwargs):
    df = data.as_dataframe()

    if "axes.facecolor" in mpl.rcParams:
        orig_facecolor = mpl.rcParams["axes.facecolor"]
        mpl.rcParams["axes.facecolor"] = "white"

    if brush is not None:
        brush_set = BrushSet(brush)
        c, color_map = color_brush(brush_set, df)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    if isinstance(x, str):
        x_label = x
        x = df[x_label]
    else:
        x_label = None

    if isinstance(y, str):
        y_label = y
        y = df[y_label]
    else:
        y_label = None

    if isinstance(z, str):
        z_label = z
        z = df[z_label]
    else:
        z_label = None

    if isinstance(c, str):
        c_label = c
        c = df[c_label]
    else:
        c_label = None

    if isinstance(s, str):
        s_label = s
        s = df[s_label]
    else:
        s_label = None

    used_keys = set([x_label, y_label, z_label, c_label, s_label, None])
    used_keys.remove(None)

    remaining_keys = list(model.responses.keys())

    for key in used_keys:
        if key in remaining_keys:
            remaining_keys.remove(key)

    for key in remaining_keys:
        if x is None:
            x_label = key
            x = df[x_label]
        elif y is None:
            y_label = key
            y = df[y_label]
        elif z is None:
            z_label = key
            z = df[z_label]
        elif c is None:
            c_label = key
            c = df[c_label]
        elif s is None:
            s_label = key
            s = df[s_label]

    if z is None:
        z = 0

    if c is None:
        c = 'b'
        show_colorbar = False

    if s is None:
        s = 20
        show_legend = False
    else:
        s_min = min(s)
        s_max = max(s)
        s = (s_range[1]-s_range[0]) * ((s-s_min) / (s_max-s_min)) + s_range[0]

    if "cmap" not in kwargs and not _is_color(c, len(x)):
        kwargs["cmap"] = RhodiumConfig.default_cmap

    handle = ax.scatter(xs=x,
                        ys=y,
                        zs=z,
                        c=c,
                        s=s,
                        picker=kwargs["picker"] if "picker" in kwargs else pick_handler is not None,
                        **kwargs)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_zlabel(z_label)

    if show_colorbar:
        if brush is None:
            cb = fig.colorbar(handle, shrink=0.5, aspect=5)
            cb.set_label(c_label)
        else:
            handle.set_array(np.asarray(color_indices(c, color_map)))
            handle.cmap = mpl.colors.ListedColormap(list(color_map.values()))
            off = (len(color_map)-1)/(len(color_map))/2
            height = (len(color_map)-1)-2*off
            ticks = [0] if len(color_map) <= 1 else [(i/(len(color_map)-1) * height + off) for i in range(len(color_map))]
            cb = fig.colorbar(handle, shrink=0.5, aspect=5, ticks=ticks)
            cb.set_label("")
            cb.ax.set_xticklabels(color_map.keys())
            cb.ax.set_yticklabels(color_map.keys())

    if show_legend:
        proxy = mpatches.Circle((0.5, 0.5), 0.25, fc="b")
        ax.legend([proxy],
                  [s_label + " (" + str(s_min) + " - " + str(s_max) + ")"],
                  handler_map={proxy: HandlerSizeLegend()})

    if interactive:
        def formatter(sel):
            i = sel.index
            point = data[i]
            keys = model.responses.keys()
            label = "Index %d" % i

            for key in keys:
                label += "\n%s: %0.2f" % (key, point[key])

            return label

        c = mplcursors.cursor(handle, hover=True)
        c.connect(
            "add", lambda sel: sel.annotation.set_text(formatter(sel)))

    if pick_handler:
        def handle_click(event):
            if hasattr(event, "ind"):
                i = event.ind[0]
                pick_handler(i)
                plt.draw()

        fig.canvas.mpl_connect('pick_event', handle_click)

    if "axes.facecolor" in mpl.rcParams:
        mpl.rcParams["axes.facecolor"] = orig_facecolor

    return fig

def scatter2d(model, data,
              x=None,
              y=None,
              c=None,
              s=None,
              s_range=(10, 50),
              show_colorbar=True,
              show_legend=False,
              interactive=False,
              brush=None,
              is_class=False,
              colors=None,
              **kwargs):
    df = data.as_dataframe()
    fig = plt.figure(facecolor='white')
    ax = plt.gca()

    if brush is not None:
        brush_set = BrushSet(brush)
        c, color_map = color_brush(brush_set, df)

    if isinstance(x, str):
        x_label = x
        x = df[x_label]
    else:
        x_label = None

    if isinstance(y, str):
        y_label = y
        y = df[y_label]
    else:
        y_label = None

    if isinstance(c, str):
        c_label = c
        c = df[c_label]
    else:
        c_label = None

    if isinstance(s, str):
        s_label = s
        s = df[s_label]
    else:
        s_label = None

    used_keys = set([x_label, y_label, c_label, s_label, None])
    used_keys.remove(None)

    remaining_keys = list(model.responses.keys())

    for key in used_keys:
        if key in remaining_keys:
            remaining_keys.remove(key)

    for key in remaining_keys:
        if x is None:
            x_label = key
            x = df[x_label]
        elif y is None:
            y_label = key
            y = df[y_label]
        elif c is None:
            c_label = key
            c = df[c_label]
        elif s is None:
            s_label = key
            s = df[s_label]

    if c is None:
        c = 'b'
        show_colorbar = False

    if s is None:
        s = 20
        show_legend = False
    else:
        s_min = min(s)
        s_max = max(s)
        s = (s_range[1]-s_range[0]) * ((s-s_min) / (s_max-s_min)) + s_range[0]

    if is_class:
        if isinstance(colors, dict):
            cmap = colors
        else:
            from pandas.plotting._matplotlib.style import get_standard_colors
            classes = c.drop_duplicates()
            color_values = get_standard_colors(num_colors=len(classes),
                                               colormap=kwargs["cmap"] if "cmap" in kwargs else None,
                                               color_type='random',
                                               color=colors)
            cmap = dict(zip(classes, color_values))
        c = [cmap[c_i] for c_i in c]
        show_colorbar = False
    
    if "cmap" not in kwargs and not _is_color(c, len(x)):
        kwargs["cmap"] = RhodiumConfig.default_cmap

    handle = plt.scatter(x=x,
                         y=y,
                         c=c,
                         s=s,
                         **kwargs)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if show_colorbar:
        if brush is None:
            cb = fig.colorbar(handle, shrink=0.5, aspect=5)
            cb.set_label(c_label)
        else:
            handle.set_array(np.asarray(color_indices(c, color_map)))
            handle.cmap = mpl.colors.ListedColormap(list(color_map.values()))
            off = (len(color_map)-1)/(len(color_map))/2
            height = (len(color_map)-1)-2*off
            ticks = [0] if len(color_map) <= 1 else [(i/(len(color_map)-1) * height + off) for i in range(len(color_map))]
            cb = fig.colorbar(handle, shrink=0.5, aspect=5, ticks=ticks)
            cb.set_label("")
            cb.ax.set_xticks(ticks)
            cb.ax.set_xticklabels(color_map.keys())
            cb.ax.set_yticklabels(color_map.keys())

    if show_legend:
        proxy = mpatches.Circle((0.5, 0.5), 0.25, fc="b")
        ax.legend([proxy],
                  [s_label + " (" + str(s_min) + " - " + str(s_max) + ")"],
                  handler_map={mpatches.Circle: HandlerSizeLegend()})

    if interactive:
        def formatter(sel):
            i = sel.index
            point = data[i]
            keys = model.responses.keys()
            label = "Index %d" % i

            for key in keys:
                label += "\n%s: %0.2f" % (key, point[key])

            return label

        c = mplcursors.cursor(handle, hover=True)
        c.connect(
            "add", lambda sel: sel.annotation.set_text(formatter(sel)))

    return fig

def joint(model, data, x, y, **kwargs):
    df = data.as_dataframe(_combine_keys(x, y))

    sns.jointplot(x=df[x],
                  y=df[y],
                  **kwargs)

def pairs(model, data,
          keys=None,
          brush=None,
          brush_label="class",
          **kwargs):
    df = data.as_dataframe(keys if keys is not None else model.responses.keys())

    if brush is None:
        sns.pairplot(df, **kwargs)
    else:
        brush_set = BrushSet(brush)
        df[brush_label] = apply_brush(brush_set, df)

        if "palette" not in kwargs:
            kwargs["palette"] = brush_color_map(brush_set, df[brush_label])

        sns.pairplot(df, hue=brush_label, **kwargs)

def kdeplot(model, data, x, y,
            brush=None,
            alpha=1.0,
            cmap=["Reds", "Blues", "Oranges", "Greens", "Greys"],
            **kwargs):
    df = data.as_dataframe()
    plt.figure()

    if brush is None:
        sns.kdeplot(x=df[x],
                    y=df[y],
                    cmap=cmap[0],
                    fill=True,
                    thresh=0.05,
                    alpha=alpha,
                    **kwargs)

        proxy = mpatches.Circle((0.5, 0.5),
                                0.25,
                                fc=sns.color_palette(cmap[0])[-2])

        ax = plt.gca()
        ax.legend([proxy], ["Density"], **kwargs)
    else:
        proxies = []
        brush_set = BrushSet(brush)

        for i, brush in reversed(list(enumerate(brush_set))):
            bin = df.query(brush.expr)
            sns.kdeplot(x=bin[x],
                        y=bin[y],
                        cmap=cmap[i % len(cmap)],
                        fill=True,
                        thresh=0.05,
                        alpha=alpha,
                        **kwargs)
            proxies.append(mpatches.Circle((0.5, 0.5),
                                           0.25,
                                           fc=sns.color_palette(cmap[i % len(cmap)])[-2]))

        ax = plt.gca()
        ax.legend(proxies, brush_set.keys(), **kwargs)

def hist(model, data, keys=None):
    df = data.as_dataframe(model.responses.keys() if keys is None else _combine_keys(keys))
    keys = model.responses.keys()

    f, axes = plt.subplots(1, len(keys))
    sns.despine(left=True)

    for i, k in enumerate(keys):
        sns.histplot(df[k], kde=False, ax=axes[i])

    plt.setp(axes, yticks=[])
    plt.tight_layout()

def interact(model, data, x, y, z, **kwargs):
    df = data.as_dataframe(_combine_keys(x, y, z))
    sns.interactplot(x, y, z, df, **kwargs)

def contour2d(model, data, x=None, y=None, z=None, levels=15, size=100, xlim=None, ylim=None, labels=True, show_colorbar=True, shrink=0.05, method='cubic', **kwargs):
    df = data.as_dataframe(_combine_keys(model.responses.keys(), x, y, z))

    if isinstance(x, str):
        x_label = x
        x = df[x_label]
    else:
        x_label = None

    if isinstance(y, str):
        y_label = y
        y = df[y_label]
    else:
        y_label = None

    if isinstance(z, str):
        z_label = z
        z = df[z_label]
    else:
        z_label = None

    used_keys = set([x_label, y_label, z_label, None])
    used_keys.remove(None)

    remaining_keys = list(model.responses.keys())

    for key in used_keys:
        if key in remaining_keys:
            remaining_keys.remove(key)

    for key in remaining_keys:
        if x is None:
            x_label = key
            x = df[x_label]
        elif y is None:
            y_label = key
            y = df[y_label]
        elif z is None:
            z_label = key
            z = df[z_label]

    # compute the grid
    if xlim is None:
        xmin = np.min(x)
        xmax = np.max(x)
    else:
        xmin = xlim[0]
        xmax = xlim[1]

    if ylim is None:
        ymin = np.min(y)
        ymax = np.max(y)
    else:
        ymin = ylim[0]
        ymax = ylim[1]

    # grid the data
    xshrink = shrink*(xmax-xmin)
    yshrink = shrink*(ymax-ymin)
    xi, yi = np.mgrid[(xmin+xshrink):(xmax-xshrink):complex(size), (ymin+yshrink):(ymax-yshrink):complex(size)]
    x = x.values
    y = y.values
    z = z.values
    x = np.reshape(x, (x.shape[0], 1))
    y = np.reshape(y, (y.shape[0], 1))
    points = np.concatenate((x, y), axis=1)

    zi = griddata(points, z, (xi, yi), method=method)

    # generate the plot
    fig = plt.figure(facecolor='white')
    ax = plt.gca()

    if labels:
        CS = plt.contour(xi, yi, zi, levels, colors="k", linewidths=0.5)
        plt.clabel(CS, **kwargs)

    if "cmap" not in kwargs:
        kwargs["cmap"] = RhodiumConfig.default_cmap

    plt.contourf(xi, yi, zi, levels, **kwargs)
    ax.set_xlim(xmin+xshrink, xmax-xshrink)
    ax.set_ylim(ymin+yshrink, ymax-yshrink)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # draw colorbar
    if show_colorbar:
        cb = plt.colorbar()
        cb.set_label(z_label)

    return fig

def contour3d(model, data, x=None, y=None, z=None, xlim=None, ylim=None, levels=15, size=100, show_colorbar=True, shrink=0.05, method='cubic', **kwargs):
    df = data.as_dataframe(_combine_keys(model.responses.keys(), x, y, z))

    if len(df.columns) < 3:
        raise ValueError("insufficient number of dimensions")

    if "axes.facecolor" in mpl.rcParams:
        orig_facecolor = mpl.rcParams["axes.facecolor"]
        mpl.rcParams["axes.facecolor"] = "white"

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    if isinstance(x, str):
        x_label = x
        x = df[x_label]
    else:
        x_label = None

    if isinstance(y, str):
        y_label = y
        y = df[y_label]
    else:
        y_label = None

    if isinstance(z, str):
        z_label = z
        z = df[z_label]
    else:
        z_label = None

    used_keys = set([x_label, y_label, z_label, None])
    used_keys.remove(None)

    remaining_keys = list(model.responses.keys())

    for key in used_keys:
        if key in remaining_keys:
            remaining_keys.remove(key)

    for key in remaining_keys:
        if x is None:
            x_label = key
            x = df[x_label]
        elif y is None:
            y_label = key
            y = df[y_label]
        elif z is None:
            z_label = key
            z = df[z_label]

    # compute the grid
    if xlim is None:
        xmin = np.min(x)
        xmax = np.max(x)
    else:
        xmin = xlim[0]
        xmax = xlim[1]

    if ylim is None:
        ymin = np.min(y)
        ymax = np.max(y)
    else:
        ymin = ylim[0]
        ymax = ylim[1]

    # grid the data
    xshrink = shrink*(xmax-xmin)
    yshrink = shrink*(ymax-ymin)
    xi, yi = np.mgrid[(xmin+xshrink):(xmax-xshrink):complex(size), (ymin+yshrink):(ymax-yshrink):complex(size)]
    x = x.values
    y = y.values
    z = z.values
    x = np.reshape(x, (x.shape[0], 1))
    y = np.reshape(y, (y.shape[0], 1))
    points = np.concatenate((x, y), axis=1)

    zi = griddata(points, z, (xi, yi), method=method)

    if "cmap" not in kwargs:
        kwargs["cmap"] = RhodiumConfig.default_cmap

    zmin = np.nanmin(zi)
    zmax = np.nanmax(zi)

    handle = ax.plot_surface(xi, yi, zi, rstride=int(size/levels), cstride=int(size/levels), vmin=zmin, vmax=zmax, **kwargs)

    if show_colorbar:
        cb = fig.colorbar(handle, shrink=0.5, aspect=5)
        cb.set_label(z_label)

    ax.set_xlim(xmin+xshrink, xmax-xshrink)
    ax.set_ylim(ymin+yshrink, ymax-yshrink)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_zlabel(z_label)

    return fig

def animate3d(prefix, dir="images/", steps=36, transform=(10, 0, 0), **kwargs):
    import os
    import math
    import inspect
    from PIL import Image

    base_dir = os.path.join(dir, prefix)

    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    ax = plt.gca()
    digits = int(math.log10(steps-1))+1
    files = []

    for n in range(steps):
        if inspect.isfunction(transform):
            transform(ax)
        else:
            ax.azim += transform[0]
            ax.elev += transform[1]
            ax.set_box_aspect(None, zoom=transform[2] if transform[2] > 0 else 1)

        filename = os.path.join(base_dir, 'img' + str(n).zfill(digits) + '.png')
        plt.savefig(filename, bbox_inches='tight')
        files.append(filename)
      
    images = [Image.open(file) for file in files]
    file_path_name = os.path.join(dir, prefix + '.gif')
    images[0].save(file_path_name, save_all=True, append_images=images[1:], **kwargs)

def parallel_coordinates(model, data, c=None, cols=None, ax=None, colors=None,
                         use_columns=False, xticks=None, colormap=None,
                         target="top", brush=None, zorder=None, **kwds):
    if "axes.facecolor" in mpl.rcParams:
        orig_facecolor = mpl.rcParams["axes.facecolor"]
        mpl.rcParams["axes.facecolor"] = "white"

    df = data.as_dataframe(_combine_keys(model.responses.keys(), cols, c))  # , exclude_dtypes=["object"])

    if brush is not None:
        brush_set = BrushSet(brush)
        assignment = apply_brush(brush_set, data)
        color_map = brush_color_map(brush_set, assignment)
        class_col = pd.DataFrame({"class": assignment})["class"]
        is_class = True
    else:
        if c is None:
            c = df.columns.values[-1]

        class_col = df[c]
        is_class = df.dtypes[c].name == "object"
        color_map = None

        if is_class:
            df = df.drop(c, axis=1)

            if c in cols:
                cols.remove(c)
        else:
            class_min = class_col.min()
            class_max = class_col.max()

    if cols is not None:
        df = df[cols]

    df_min = df.min()
    df_max = df.max()

    df = (df - df_min) / (df_max - df_min)
    n = len(df)

    used_legends = set([])

    ncols = len(df.columns)

    for i in range(ncols):
        if target == "top":
            if model.responses[df.columns.values[i]].dir == Response.MINIMIZE:
                df.iloc[:, i] = 1-df.iloc[:, i]
        elif target == "bottom":
            if model.responses[df.columns.values[i]].dir == Response.MAXIMIZE:
                df.iloc[:, i] = 1-df.iloc[:, i]

    # determine values to use for xticks
    if use_columns is True:
        if not np.all(np.isreal(list(df.columns))):
            raise ValueError('Columns must be numeric to be used as xticks')
        x = df.columns
    elif xticks is not None:
        if not np.all(np.isreal(xticks)):
            raise ValueError('xticks specified must be numeric')
        elif len(xticks) != ncols:
            raise ValueError('Length of xticks must match number of columns')
        x = xticks
    else:
        x = range(ncols)

    if ax is None:
        fig = plt.figure()
        ax = plt.gca()
    else:
        fig = ax.get_figure()

    cmap = plt.get_cmap(colormap)

    if is_class:
        if color_map is None:
            if isinstance(colors, dict):
                cmap = colors
            else:
                from pandas.plotting._matplotlib.style import get_standard_colors
                classes = class_col.drop_duplicates()
                color_values = get_standard_colors(num_colors=len(classes),
                                                   colormap=colormap,
                                                   color_type='random',
                                                   color=colors)
                cmap = dict(zip(classes, color_values))
        else:
            cmap = color_map

    if zorder is None:
        indices = range(n)
    else:
        indices = [i[0] for i in sorted(enumerate(df[zorder]), key=lambda x: x[1])]

    for i in indices:
        y = df.iloc[i].values
        kls = class_col.iat[i]

        if is_class:
            label = str(kls)

            if label not in used_legends:
                used_legends.add(label)
                ax.plot(x, y, label=label, color=cmap[kls], **kwds)
            else:
                ax.plot(x, y, color=cmap[kls], **kwds)
        else:
            ax.plot(x, y, color=cmap((kls - class_min)/(class_max-class_min)), **kwds)

    for i in x:
        ax.axvline(i, linewidth=2, color='black')
        format = "%.2f"

        if target == "top":
            value = df_min.iloc[i] if model.responses[df.columns.values[i]].dir == Response.MINIMIZE else df_max.iloc[i]

            if model.responses[df.columns.values[i]].dir != Response.INFO:
                format = format + "*"
        elif target == "bottom":
            value = df_max.iloc[i] if model.responses[df.columns.values[i]].dir == Response.MINIMIZE else df_min.iloc[i]
        else:
            value = df_max.iloc[i]

            if model.responses[df.columns.values[i]].dir == Response.MAXIMIZE:
                format = format + "*"

        ax.text(i, 1.001, format % value, ha="center", fontsize=10)
        format = "%.2f"

        if target == "top":
            value = df_max.iloc[i] if model.responses[df.columns.values[i]].dir == Response.MINIMIZE else df_min.iloc[i]
        elif target == "bottom":
            value = df_min.iloc[i] if model.responses[df.columns.values[i]].dir == Response.MINIMIZE else df_max.iloc[i]

            if model.responses[df.columns.values[i]].dir != Response.INFO:
                format = format + "*"
        else:
            value = df_min.iloc[i]

            if model.responses[df.columns.values[i]].dir == Response.MINIMIZE:
                format = format + "*"

        ax.text(i, -0.001, format % value, ha="center", va="top", fontsize=10)

    ax.set_yticks([])
    ax.set_xticks(x)
    ax.set_xticklabels(df.columns)
    ax.set_xlim(x[0]-0.1, x[-1]+0.1)
    ax.tick_params(direction="out", pad=10)

    bbox_props = dict(boxstyle="rarrow,pad=0.3", fc="white", ec="black", lw=2)
    if target == "top":
        ax.text(-0.05, 0.5, "Target", ha="center", va="center", rotation=90, bbox=bbox_props, transform=ax.transAxes)
    elif target == "bottom":
        ax.text(-0.05, 0.5, "Target", ha="center", va="center", rotation=-90, bbox=bbox_props, transform=ax.transAxes)

    if is_class:
        ax.legend(loc='center right', bbox_to_anchor=(1.25, 0.5))
        fig.subplots_adjust(right=0.8)
    else:
        cax, _ = mpl.colorbar.make_axes(ax)
        cb = mpl.colorbar.ColorbarBase(cax, cmap=cmap, spacing='proportional', norm=mpl.colors.Normalize(vmin=class_min, vmax=class_max), format='%.2f')
        cb.set_label(c)
        cb.set_clim(class_min, class_max)

    mpl.rcParams["axes.facecolor"] = orig_facecolor

    return fig



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import math
import inspect
import operator
import itertools
from SALib.sample import saltelli, fast_sampler, finite_diff, latin
from SALib.analyze import sobol, fast, dgsm, delta
from SALib.sample import morris as morris_sampler
from SALib.analyze import morris as morris_analyzer
from SALib.sample import ff as ff_sampler
from SALib.analyze import ff as ff_analyzer
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from .model import overwrite
from .optimization import evaluate
from .sampling import sample_lhs

def _cleanup_kwargs(function, kwargs):
    argspec = inspect.getfullargspec(function)
    result = {}

    if not argspec.varkw:
        for key in kwargs.keys():
            if key in argspec.args or hasattr(argspec, "kwargs"):
                result[key] = kwargs[key]

    return result

def _S2_to_dict(matrix, problem):
    result = {}
    names = list(problem["names"])

    for i in range(problem["num_vars"]):
        for j in range(i+1, problem["num_vars"]):
            if names[i] not in result:
                result[names[i]] = {}
            if names[j] not in result:
                result[names[j]] = {}

            result[names[i]][names[j]] = result[names[j]][names[i]] = float(matrix[i][j])

    return result

def _predict_N(method, nsamples, nvars, kwargs):
    if method == "sobol":
        if kwargs.get("calc_second_order", True):
            return int(math.ceil(nsamples / (2*nvars+2)))
        else:
            return int(math.ceil(nsamples / (nvars+2)))
    elif method == "morris":
        return int(math.ceil(nsamples / (nvars+1)))
    elif method == "fast":
        return int(math.ceil(nsamples / nvars))
    elif method == "dgsm":
        return int(math.ceil(nsamples / (nvars+1)))
    elif method == "delta":
        return nsamples
    else:
        return nsamples

class SAResult(dict):

    def __init__(self, parameters, *args, **kwargs):
        super(SAResult, self).__init__(*args, **kwargs)
        self.parameters = parameters

    def _longest_name(self):
        return max(len(k) for k in self.parameters)

    def __str__(self):
        lines = []
        format_str = "    %" + str(self._longest_name()) + "s: %+8f"
        format_conf = " (%+8f)"
        format_second = "    %" + str(2*self._longest_name()+3) + "s: %+8f"

        if "S1" in self:
            lines.append("First order sensitivity indices (confidence interval):")
            for k in self.parameters:
                line = format_str % (k, self["S1"][k])
                if "S1_conf" in self:
                    line += format_conf % self["S1_conf"][k]
                lines.append(line)

        if "ST" in self:
            lines.append("Total order sensitivity indices (confidence interval):")
            for k in self.parameters:
                line = format_str % (k, self["ST"][k])
                if "S1_conf" in self:
                    line += format_conf % self["ST_conf"][k]
                lines.append(line)

        if "S2" in self:
            lines.append("Second order sensitivity indices (confidence interval):")
            for k1, k2 in itertools.combinations(self.parameters, 2):
                line = format_second % (k1 + " - " + k2, self["S2"][k1][k2])
                if "S2_conf" in self:
                    line += format_conf % self["S2_conf"][k1][k2]
                lines.append(line)

        if "delta" in self:
            lines.append("Borgonovo's delta moment (confidence interval):")
            for k in self.parameters:
                line = format_str % (k, self["delta"][k])
                if "delta_conf" in self:
                    line += format_conf % self["delta_conf"][k]
                lines.append(line)

        if "vi" in self:
            lines.append("DGSM's Importance Criteria (stdev):")
            for k in self.parameters:
                line = format_str % (k, self["vi"][k])
                if "vi_std" in self:
                    line += format_conf % self["vi_std"][k]
                lines.append(line)

        if "dgsm" in self:
            lines.append("DGSM's Sensitivity Index (confidence interval):")
            for k in self.parameters:
                line = format_str % (k, self["dgsm"][k])
                if "dgsm_conf" in self:
                    line += format_conf % self["dgsm_conf"][k]
                lines.append(line)

        if "mu" in self:
            lines.append("Morris Method's mu:")
            for k in self.parameters:
                lines.append(format_str % (k, self["mu"][k]))

        if "mu_star" in self:
            lines.append("Morris Method's mu* (confidence interval):")
            for k in self.parameters:
                line = format_str % (k, self["mu_star"][k])
                if "mu_star_conf" in self:
                    line += format_conf % self["mu_star_conf"][k]
                lines.append(line)

        if "sigma" in self:
            lines.append("Morris Method's sigma:")
            for k in self.parameters:
                lines.append(format_str % (k, self["sigma"][k]))

        return "\n".join(lines)

    def plot(self, **kwargs):
        nfigs = sum([k in ("S1", "ST", "delta") for k in self.keys()])
        fig, axarr = plt.subplots(1, nfigs, sharey=True)
        n = len(self.parameters)
        index = 0

        if "S1" in self:
            axarr[index].bar(range(n),
                             [self["S1"][k] for k in self.parameters],
                             yerr=[self["S1_conf"][k] for k in self.parameters] if "S1_conf" in self else None,
                             align="center")
            axarr[index].set_title("First order sensitivity indices")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "ST" in self:
            axarr[index].bar(range(n),
                             [self["ST"][k] for k in self.parameters],
                             yerr=[self["ST_conf"][k] for k in self.parameters] if "ST_conf" in self else None,
                             align="center")
            axarr[index].set_title("Total order sensitivity indices")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "delta" in self:
            axarr[index].bar(range(n),
                             [self["delta"][k] for k in self.parameters],
                             yerr=[self["delta_conf"][k] for k in self.parameters] if "delta_conf" in self else None,
                             align="center")
            axarr[index].set_title("Borgonovo's delta moment")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "vi" in self:
            axarr[index].bar(range(n),
                             [self["vi"][k] for k in self.parameters],
                             yerr=[self["vi_std"][k] for k in self.parameters] if "vi_std" in self else None,
                             align="center")
            axarr[index].set_title("DGSM's Importance Criteria")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "dgsm" in self:
            axarr[index].bar(range(n),
                             [self["dgsm"][k] for k in self.parameters],
                             yerr=[self["dgsm_conf"][k] for k in self.parameters] if "dgsm_conf" in self else None,
                             align="center")
            axarr[index].set_title("DGSM's Sensitivity Index")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "mu" in self:
            axarr[index].bar(range(n),
                             [self["mu"][k] for k in self.parameters],
                             align="center")
            axarr[index].set_title("Morris Method's mu")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "mu_star" in self:
            axarr[index].bar(range(n),
                             [self["mu_star"][k] for k in self.parameters],
                             yerr=[self["mu_star_conf"][k] for k in self.parameters] if "mu_star_conf" in self else None,
                             align="center")
            axarr[index].set_title("Morris Method's mu*")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        if "sigma" in self:
            axarr[index].bar(range(n),
                             [self["sigma"][k] for k in self.parameters],
                             align="center")
            axarr[index].set_title("Morris Method's sigma")
            axarr[index].set_xticks(range(n))
            axarr[index].set_xticklabels(self.parameters)
            index += 1

        return fig

    def _is_significant(self, value, confidence_interval, threshold="conf"):
        if threshold == "conf":
            return value - abs(confidence_interval) > 0
        else:
            return value - abs(float(threshold)) > 0

    def plot_sobol(self, radSc=2.0, scaling=1, widthSc=0.5, STthick=1, varNameMult=1.3, colors=None, groups=None, gpNameMult=1.5, threshold="conf"):
        # Derived from https://github.com/calvinwhealton/SensitivityAnalysisPlots
        fig, ax = plt.subplots(1, 1)
        color_map = {}

        # initialize parameters and colors
        if groups is None:
            parameters = list(self.parameters)

            if colors is None:
                colors = ["k"]

            for i, parameter in enumerate(parameters):
                color_map[parameter] = colors[i % len(colors)]
        else:
            parameters = []

            if colors is None:
                colors = sns.color_palette("deep", max(3, len(groups)))

            for i, key in enumerate(groups.keys()):
                parameters.extend(groups[key])

                for parameter in groups[key]:
                    color_map[parameter] = colors[i % len(colors)]

        n = len(parameters)
        angles = radSc*math.pi*np.arange(0, n)/n
        x = radSc*np.cos(angles)
        y = radSc*np.sin(angles)

        # plot second-order indices
        for i, j in itertools.combinations(range(n), 2):
            key1 = parameters[i]
            key2 = parameters[j]

            if self._is_significant(self["S2"][key1][key2], self["S2_conf"][key1][key2], threshold):
                angle = math.atan((y[j]-y[i])/(x[j]-x[i]))

                if y[j]-y[i] < 0:
                    angle += math.pi

                line_hw = scaling*(max(0, self["S2"][key1][key2])**widthSc)/2

                coords = np.empty((4, 2))
                coords[0, 0] = x[i] - line_hw*math.sin(angle)
                coords[1, 0] = x[i] + line_hw*math.sin(angle)
                coords[2, 0] = x[j] + line_hw*math.sin(angle)
                coords[3, 0] = x[j] - line_hw*math.sin(angle)
                coords[0, 1] = y[i] + line_hw*math.cos(angle)
                coords[1, 1] = y[i] - line_hw*math.cos(angle)
                coords[2, 1] = y[j] - line_hw*math.cos(angle)
                coords[3, 1] = y[j] + line_hw*math.cos(angle)

                ax.add_artist(plt.Polygon(coords, color="0.75"))

        # plot total order indices
        for i, key in enumerate(parameters):
            if self._is_significant(self["ST"][key], self["ST_conf"][key], threshold):
                ax.add_artist(plt.Circle((x[i], y[i]), scaling*(self["ST"][key]**widthSc)/2, color='w'))
                ax.add_artist(plt.Circle((x[i], y[i]), scaling*(self["ST"][key]**widthSc)/2, lw=STthick, color='0.4', fill=False))

        # plot first-order indices
        for i, key in enumerate(parameters):
            if self._is_significant(self["S1"][key], self["S1_conf"][key], threshold):
                ax.add_artist(plt.Circle((x[i], y[i]), scaling*(self["S1"][key]**widthSc)/2, color='0.4'))

        # add labels
        for i, key in enumerate(parameters):
            ax.text(varNameMult*x[i], varNameMult*y[i], key, ha='center', va='center',
                    rotation=angles[i]*360/(2*math.pi) - 90,
                    color=color_map[key])

        if groups is not None:
            for i, group in enumerate(groups.keys()):
                group_angle = np.mean([angles[j] for j in range(n) if parameters[j] in groups[group]])

                ax.text(gpNameMult*radSc*math.cos(group_angle),
                        gpNameMult*radSc*math.sin(group_angle), group, ha='center', va='center',
                        rotation=group_angle*360/(2*math.pi) - 90,
                        color=colors[i % len(colors)])

        ax.set_facecolor('white')
        ax.set_xticks([])
        ax.set_yticks([])
        plt.axis('equal')
        plt.axis([-2*radSc, 2*radSc, -2*radSc, 2*radSc])

        return fig

def sa(model, response, policy={}, method="sobol", nsamples=1000, **kwargs):
    if len(model.uncertainties) == 0:
        raise ValueError("no uncertainties defined in model")

    problem = {'num_vars': len(model.uncertainties),
               'names': model.uncertainties.keys(),
               'bounds': [[0.0, 1.0] for u in model.uncertainties],
               'groups': kwargs.get("groups", None)}

    # estimate the argument N passed to the sampler that produces the requested
    # number of samples
    N = _predict_N(method, nsamples, problem["num_vars"], kwargs)

    # generate the samples
    if method == "sobol":
        samples = saltelli.sample(problem, N, **_cleanup_kwargs(saltelli.sample, kwargs))
    elif method == "morris":
        samples = morris_sampler.sample(problem, N, **_cleanup_kwargs(morris_sampler.sample, kwargs))
    elif method == "fast":
        samples = fast_sampler.sample(problem, N, **_cleanup_kwargs(fast_sampler.sample, kwargs))
    elif method == "ff":
        samples = ff_sampler.sample(problem, **_cleanup_kwargs(ff_sampler.sample, kwargs))
    elif method == "dgsm":
        samples = finite_diff.sample(problem, N, **_cleanup_kwargs(finite_diff.sample, kwargs))
    elif method == "delta":
        if "samples" in kwargs:
            samples = kwargs["samples"]
        else:
            samples = latin.sample(problem, N, **_cleanup_kwargs(latin.sample, kwargs))

    # convert from samples in [0, 1] to uncertainty domain
    for i, u in enumerate(model.uncertainties):
        samples[:, i] = u.ppf(samples[:, i])

    # run the model and collect the responses
    responses = np.empty(samples.shape[0])

    for i in range(samples.shape[0]):
        sample = {k: v for k, v in zip(model.uncertainties.keys(), samples[i])}
        responses[i] = evaluate(model, overwrite(sample, policy))[response]

    # run the sensitivity analysis method
    if method == "sobol":
        result = sobol.analyze(problem, responses, **_cleanup_kwargs(sobol.analyze, kwargs))
    elif method == "morris":
        result = morris_analyzer.analyze(problem, samples, responses, **_cleanup_kwargs(morris_analyzer.analyze, kwargs))
    elif method == "fast":
        result = fast.analyze(problem, responses, **_cleanup_kwargs(fast.analyze, kwargs))
    elif method == "ff":
        result = ff_analyzer.analyze(problem, samples, responses, **_cleanup_kwargs(ff_analyzer.analyze, kwargs))
    elif method == "dgsm":
        result = dgsm.analyze(problem, samples, responses, **_cleanup_kwargs(dgsm.analyze, kwargs))
    elif method == "delta":
        result = delta.analyze(problem, samples, responses, **_cleanup_kwargs(delta.analyze, kwargs))

    # convert the SALib results into a form allowing pretty printing and
    # lookups using the parameter name
    pretty_result = SAResult(list(result["names"] if "names" in result else problem["names"]))

    if "S1" in result:
        pretty_result["S1"] = {k: float(v) for k, v in zip(problem["names"], result["S1"])}
    if "S1_conf" in result:
        pretty_result["S1_conf"] = {k: float(v) for k, v in zip(problem["names"], result["S1_conf"])}
    if "ST" in result:
        pretty_result["ST"] = {k: float(v) for k, v in zip(problem["names"], result["ST"])}
    if "ST_conf" in result:
        pretty_result["ST_conf"] = {k: float(v) for k, v in zip(problem["names"], result["ST_conf"])}
    if "S2" in result:
        pretty_result["S2"] = _S2_to_dict(result["S2"], problem)
    if "S2_conf" in result:
        pretty_result["S2_conf"] = _S2_to_dict(result["S2_conf"], problem)
    if "delta" in result:
        pretty_result["delta"] = {k: float(v) for k, v in zip(problem["names"], result["delta"])}
    if "delta_conf" in result:
        pretty_result["delta_conf"] = {k: float(v) for k, v in zip(problem["names"], result["delta_conf"])}
    if "vi" in result:
        pretty_result["vi"] = {k: float(v) for k, v in zip(problem["names"], result["vi"])}
    if "vi_std" in result:
        pretty_result["vi_std"] = {k: float(v) for k, v in zip(problem["names"], result["vi_std"])}
    if "dgsm" in result:
        pretty_result["dgsm"] = {k: float(v) for k, v in zip(problem["names"], result["dgsm"])}
    if "dgsm_conf" in result:
        pretty_result["dgsm_conf"] = {k: float(v) for k, v in zip(problem["names"], result["dgsm_conf"])}
    if "mu" in result:
        pretty_result["mu"] = {k: float(v) for k, v in zip(result["names"], result["mu"])}
    if "mu_star" in result:
        pretty_result["mu_star"] = {k: float(v) for k, v in zip(result["names"], result["mu_star"])}
    if "mu_star_conf" in result:
        pretty_result["mu_star_conf"] = {k: float(v) for k, v in zip(result["names"], result["mu_star_conf"])}
    if "sigma" in result:
        pretty_result["sigma"] = {k: float(v) for k, v in zip(result["names"], result["sigma"])}

    return pretty_result

def oat(model, response, policy={}, nsamples=100, **kwargs):
    keys = model.uncertainties.keys()
    responses = np.empty((nsamples, len(keys)))
    samples = np.linspace(0.1, 0.99, num=nsamples)
    base_response = evaluate(model, policy)[response]

    for i, u in enumerate(model.uncertainties):
        ppf_samples = u.ppf(samples)

        for j in range(nsamples):
            sample = {u.name: ppf_samples[j]}
            responses[j, i] = evaluate(model, overwrite(sample, policy))[response]

    minv = np.nanmin(responses, axis=0)
    maxv = np.nanmax(responses, axis=0)
    midv = responses[nsamples//2]

    total_range = maxv-minv
    negative_percentage = abs(midv - minv) / total_range
    positive_percentage = abs(maxv - midv) / total_range

    total_range /= np.max(total_range)
    negative_percentage *= total_range / 2
    positive_percentage *= total_range / 2

    fig, (ax1, ax2) = plt.subplots(1, 2)

    # draw the lines first to get colors
    colors = []

    for i, u in enumerate(model.uncertainties):
        h = ax2.plot(samples, responses[:, i])
        colors.append(h[0].get_color())

    ax2.axvline(0.5, ls='--', color='k')
    ax2.axhline(base_response, ls='--', color='k')
    ax2.set_xlim([0.1, 0.99])
    ax2.set_xticks([0.1, 0.25, 0.5, 0.75, 0.99])
    ax2.set_xticklabels(["1", "25", "50", "75", "99"])
    ax2.set_xlabel("Quantile of Prior (%)")
    ax2.set_ylabel(response)
    ax2.yaxis.tick_right()
    ax2.yaxis.set_label_position("right")
    ax2.legend(keys)

    # then draw the tornado chart
    ax1.barh(range(len(keys)), negative_percentage, left=0.5-negative_percentage, align='center', color=colors)
    ax1.barh(range(len(keys)), positive_percentage, left=0.5, align='center', color=colors)
    ax1.set_yticks(range(len(keys)))
    ax1.set_yticklabels(keys)
    ax1.set_xticks([0.0, 0.25, 0.5, 0.75, 1.0])
    ax1.set_xticklabels(["-100%", "-50%", "0%", "50%", "100%"])
    ax1.set_xlabel("Percent of Total Variance")
    ax1.set_xlim([0, 1])

    return fig

def regional_sa(model, expr, policy={}, nsamples=1000):
    samples = sample_lhs(model, nsamples)
    output = evaluate(model, overwrite(samples, policy))
    classification = output.apply(expr)
    classes = sorted(set(classification))
    fig, axarr = plt.subplots(1, len(model.uncertainties))
    lines = []

    for i, u in enumerate(model.uncertainties):
        for k in classes:
            indices = [classification[j] == k for j in range(len(classification))]
            values = [output[j][u.name] for j in range(len(indices)) if indices[j]]
            sorted_values = sorted(enumerate(values), key=operator.itemgetter(1))
            h = axarr[i].plot([v[1] for v in sorted_values], np.arange(len(values))/float(len(values)-1))
            lines.append(h[0])

        values = [output[j][u.name] for j in range(len(indices))]
        sorted_values = sorted(enumerate(values), key=operator.itemgetter(1))
        h = axarr[i].plot([v[1] for v in sorted_values], np.arange(len(values))/float(len(values)-1))
        lines.append(h[0])

        axarr[i].set_title(u.name)

    plt.figlegend(lines[:len(classes)] + [lines[-1]],
                  list(map(str, classes)) + ["Unconditioned"],
                  loc='lower center',
                  ncol=3,
                  labelspacing=0.0)

    return fig



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import win32com.client
from win32com.universal import com_error
from .model import Model

class ExcelHelper(object):

    def __init__(self, filename, sheet=1, visible=False):
        super(ExcelHelper, self).__init__()
        self.xl = win32com.client.Dispatch("Excel.Application")
        self.wb = self.xl.Workbooks.Open(filename)

        # ensure auto-calculations is enabled
        sheets = self.xl.Worksheets

        for i in range(sheets.Count):
            sheets.Item(i+1).EnableCalculation = True

        # set the default sheet
        self.set_sheet(sheet)

        if visible:
            self.show()

    def set_sheet(self, sheet):
        try:
            self.sheet = self.wb.Sheets(sheet)
            self.sheet_index = sheet
        except com_error:
            raise ValueError("invalid sheet " + str(sheet))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def show(self):
        self.xl.Visible = 1

    def hide(self):
        self.xl.Visible = 0

    def close(self):
        if self.wb:
            self.wb.Close(SaveChanges=0)
            self.wb = None

        if self.xl:
            self.xl.DisplayAlerts = False
            self.xl.Quit()
            self.xl = None

    def _dereference(self, cell):
        try:
            if isinstance(cell, str):
                return self.sheet.Range(cell)
            elif isinstance(cell, (list, tuple)) and len(cell) == 2:
                return self.sheet.Cells(cell[0], cell[1])
        except com_error:
            pass

        raise ValueError("invalid cell reference " + str(cell))

    def __getitem__(self, cell):
        return self._dereference(cell).Value

    def __setitem__(self, cell, value):
        # expects single value or 2D list/tuple
        if isinstance(value, (list, tuple)):
            value = [v if isinstance(v, (list, tuple)) else [v] for v in value]

        self._dereference(cell).Value = value

class ExcelModel(Model):

    def __init__(self, filename, **kwargs):
        super(ExcelModel, self).__init__(self._evaluate)
        self.excel_helper = ExcelHelper(filename, **kwargs)

    def _evaluate(self, **kwargs):
        result = {}

        for parameter in self.parameters:
            if hasattr(parameter, "cell"):
                key = getattr(parameter, "cell")
            else:
                key = parameter.name

            if hasattr(parameter, "sheet") and self.excel_helper.sheet_index != getattr(parameter, "sheet"):
                self.excel_helper.set_sheet(getattr(parameter, "sheet"))

            value = kwargs.get(parameter.name, parameter.default_value)

            if value is not None:
                self.excel_helper[key] = value

        for response in self.responses:
            if hasattr(response, "cell"):
                key = getattr(response, "cell")
            else:
                key = response.name

            if hasattr(response, "sheet") and self.excel_helper.sheet_index != getattr(response, "sheet"):
                self.excel_helper.set_sheet(getattr(response, "sheet"))

            result[response.name] = self.excel_helper[key]

        return result

    def close(self):
        self.excel_helper.close()
        super(ExcelModel, self).close()



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from rhodium import Model, UniformUncertainty, sample_lhs, sample_uniform

class TestSampling(unittest.TestCase):

    def testUniform(self):
        model = Model("foo")
        model.uncertainties = [UniformUncertainty("x", 5.0, 10.0)]

        samples = sample_uniform(model, 100)

        self.assertEqual(100, len(samples))

        for i in range(len(samples)):
            self.assertTrue("x" in samples[i])
            self.assertTrue(samples[i]["x"] >= 5.0 and samples[i]["x"] <= 10.0)

    def testLHS(self):
        model = Model("foo")
        model.uncertainties = [UniformUncertainty("x", 5.0, 10.0)]

        samples = sample_lhs(model, 100)

        self.assertEqual(100, len(samples))

        for i in range(len(samples)):
            self.assertTrue("x" in samples[i])
            self.assertTrue(samples[i]["x"] >= 5.0 and samples[i]["x"] <= 10.0)



import os
import time
import subprocess

devnull = open(os.devnull, "w")
cwd = os.getcwd()
examples = ["Python", "C", "R", "Excel"]

for example in examples:
    os.chdir(os.path.join(cwd, example))
    start_time = time.time()
    subprocess.call(["python", "lakeModelIn" + example + ".py"], stdout=devnull)
    print("Lake Model in " + example + ": " + str(time.time() - start_time))



import math
import numpy as np
from scipy.optimize import brentq as root
from rhodium import *

# Define the lake problem
def lake_problem(
         pollution_limit, # array of 100 values specifying the pollution limit each time period
         b = 0.42,        # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,         # recycling exponent
         mean = 0.02,     # mean of natural inflows
         stdev = 0.001,   # standard deviation of natural inflows
         alpha = 0.4,     # utility from pollution
         delta = 0.98,    # future utility discount rate
         nsamples = 100): # monte carlo sampling of natural inflows
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(pollution_limit)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(pollution_limit)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)

        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)

    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    intertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return (max_P, utility, intertia, reliability)

# Construct the Rhodium model using the Python function
model = Model(lake_problem)

# List all the parameters.  The names must match the Python function arguments.
model.parameters = [Parameter("pollution_limit"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

# List all the outputs.  Since the Python function is returning a tuple/list, the
# order matters.  We could also return a dict.
model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

# Specify the levers
model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

# Optimize the model using Rhodium
output = optimize(model, "NSGAII", 10000)
print(output)


# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from ..cache import cached, cache, setup_cache

class TestCache(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        setup_cache(clear=True)

    def testSimpleCache(self):
        cache("testValue", 5)
        value = cache("testValue", 10)  # Should return the cached value of 5
        self.assertEqual(5, value)

    def testFunction(self):
        self.testFunction_called = False

        def fun():
            self.testFunction_called = True
            return 5

        value = cache("testFun", fun)
        self.assertEqual(True, self.testFunction_called)
        self.assertEqual(5, value)
        self.testFunction_called = False

        value = cache("testFun", fun)
        self.assertEqual(False, self.testFunction_called)
        self.assertEqual(5, value)

    def testCachedFunction(self):
        self.testCachedFunction_called = False

        @cached
        def fun():
            self.testCachedFunction_called = True
            return 5

        value = fun()
        self.assertEqual(True, self.testCachedFunction_called)
        self.assertEqual(5, value)
        self.testCachedFunction_called = False

        value = fun()
        self.assertEqual(False, self.testCachedFunction_called)
        self.assertEqual(5, value)



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import os
import ast
import math
import inspect
import random
import operator
import pandas as pd
import scipy.stats as stats
from collections import OrderedDict
from abc import ABCMeta, abstractmethod
from platypus import Real, Integer, Permutation, Subset
from .expr import _evaluate_all

class RhodiumError(Exception):
    pass

class NamedObject(object):
    """Object with a name."""

    def __init__(self, name):
        super(NamedObject, self).__init__()
        self.name = name

class Parameter(NamedObject):
    """Defines a model parameter (i.e., input).

    Defines a model input (i.e., input) and an optional default value.  The
    name must be identical to the keyword argument to the method defining the
    model.
    """

    def __init__(self, name, default_value=None, **kwargs):
        super(Parameter, self).__init__(name)
        self.default_value = default_value

        for k, v in kwargs.items():
            setattr(self, k, v)

class Response(NamedObject):
    """Defines a model response (i.e., output).

    Defines a model response (i.e., output) and its type.  The type can be
    MINIMIZE, MAXIMIZE, or INFO.  If MINIMIZE or MAXIMIZE, then the response
    may be used during optimization.  If INFO, the default, the response is
    purely for informative purposes (e.g., for generating plots) but does not
    participate in optimization.
    """

    MINIMIZE = -1
    MAXIMIZE = 1
    INFO = 2
    IGNORE = 0

    def __init__(self, name, dir=INFO, **kwargs):
        super(Response, self).__init__(name)
        self.dir = dir

        for k, v in kwargs.items():
            setattr(self, k, v)

_eval_env = {}
module = __import__("math", fromlist=[''])

for name in dir(module):
    if not name.startswith("_"):
        _eval_env[name] = getattr(module, name)

class Constraint(object):
    """Defines model constraints.

    Defines constraints that must be satisfied in order for a policy to be
    considered feasible.  This is often called a "hard constraint."
    Constraints can either be defined using a valid Python expression that
    references any parameters or responses (passed as a string) or a function
    given a dict of the parameters and responses.
    """

    def __init__(self, expr):
        super(Constraint, self).__init__()
        self.expr = expr

        if isinstance(expr, str):
            self._convert()

    def _convert(self):
        """Attempts to convert expression to distance function.

        Constraints are often expressed as inequalities, such as x < 5, meaning
        that a policy is feasible if the value of x is less than 5.  It is
        sometimes useful to know how far a policy is from a feasibility
        threshold.  For example, x = 7 is closer to the feasibility threshold
        than x = 15.

        This method attempts to convert a comparison expression to a distance
        expression by manipulating the AST.  If successful, this method creates
        the _distance attribute.  Even if this method is successful, the
        generated expression may not be valid.
        """
        root = ast.parse(self.expr, mode="eval")

        if isinstance(root.body, ast.Compare) and len(root.body.ops) == 1:
            left_expr = root.body.left
            right_expr = root.body.comparators[0]

            distance_expr = ast.Expression(ast.BinOp(left_expr,
                                                     ast.Sub(),
                                                     right_expr))

            ast.fix_missing_locations(distance_expr)
            self._distance = compile(distance_expr, "<AST>", "eval")

    def is_feasible(self, env):
        """Returns True if the constraint is feasible / satisfied, otherwise False."""
        tmp_env = {}
        tmp_env.update(_eval_env)
        tmp_env.update(env)

        if isinstance(self.expr, str):
            return eval(self.expr, {}, tmp_env)
        else:
            return self.expr(tmp_env)

    def distance(self, env):
        """Returns the distance to the feasibility threshold."""
        if self.is_feasible(env):
            return 0.0
        elif hasattr(self, "_distance"):
            try:
                tmp_env = {}
                tmp_env.update(_eval_env)
                tmp_env.update(env)
                return abs(eval(self._distance, {}, tmp_env)) + 0.001
            except Exception:
                return 1.0
        else:
            return 1.0

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_distance"]
        return state

    def __setstate__(self, newstate):
        self.__dict__.update(newstate)

        if isinstance(self.expr, str):
            self._convert()

class Lever(NamedObject):
    """Defines an adjustable lever that controls a model parameter.

    Model parameters can either be constant, controlled by a lever, or
    subject to uncertainty.  The lever defines the available options for
    a given design factor.

    All levers must define a length attribute, which specifies the number of
    decision variables required to represent this lever in Platypus.
    """

    __metaclass__ = ABCMeta

    def __init__(self, name):
        super(Lever, self).__init__(name)

    @abstractmethod
    def to_variables(self):
        raise NotImplementedError("method not implemented")

    @abstractmethod
    def from_variables(self, variables):
        raise NotImplementedError("method not implemented")

class RealLever(Lever):
    """Defines a lever for real values."""

    def __init__(self, name, min_value, max_value, length=1):
        super(RealLever, self).__init__(name)
        self.min_value = float(min_value)
        self.max_value = float(max_value)
        self.length = length

    def to_variables(self):
        return [Real(self.min_value, self.max_value) for _ in range(self.length)]

    def from_variables(self, variables):
        if self.length == 1:
            return variables[0]
        else:
            return variables

class IntegerLever(Lever):
    """Defines a lever for integer values."""

    def __init__(self, name, min_value, max_value, length=1):
        super(IntegerLever, self).__init__(name)
        self.min_value = int(min_value)
        self.max_value = int(max_value)
        self.length = length

    def to_variables(self):
        return [Integer(self.min_value, self.max_value) for _ in range(self.length)]

    def from_variables(self, variables):
        if self.length == 1:
            return variables[0]
        else:
            return variables

class CategoricalLever(Lever):
    """Defines a lever for categorical values (i.e., an enumeration of distinct values)."""

    def __init__(self, name, categories):
        super(CategoricalLever, self).__init__(name)
        self.categories = list(categories)
        self.length = 1

    def to_variables(self):
        return [Integer(0, len(self.categories)-1)]

    def from_variables(self, variables):
        return self.categories[variables[0]]

class PermutationLever(Lever):
    """Defines a lever for a permutation of values."""

    def __init__(self, name, options):
        super(PermutationLever, self).__init__(name)
        self.options = list(options)
        self.length = 1

    def to_variables(self):
        return [Permutation(self.options)]

    def from_variables(self, variables):
        return variables[0]

class SubsetLever(Lever):
    """Defines a lever for a fixed-size subset of a set of values."""

    def __init__(self, name, options, size):
        super(SubsetLever, self).__init__(name)
        self.options = list(options)
        self.size = size
        self.length = 1

    def to_variables(self):
        return [Subset(self.options, self.size)]

    def from_variables(self, variables):
        return variables[0]

class Uncertainty(NamedObject):
    """Defines an uncertainty for a model parameter.

    An uncertainty indicates a model parameter falls within a given
    distribution.  The specific subclass defines the distribution.
    """

    __metaclass__ = ABCMeta

    def __init__(self, name):
        super(Uncertainty, self).__init__(name)

    @abstractmethod
    def levels(self, nlevels):
        """Returns a random sampling from the uncertainty distribution in n levels.

        Used by Latin hypercube sampling, where a sample is taken from n levels
        across the distribution.  For example, a uniform distribution with 3 levels
        would result in three random values in the range [0-1/3], [1/3-2/3], [2/3-1].
        For non-uniform distributions, the levels are proportional to the PPF of the
        distribution."""
        raise NotImplementedError("method not implemented")

    @abstractmethod
    def ppf(self, x):
        """The Percent Point Function, or inverse of the Cumulative Distribution Function."""
        raise NotImplementedError("method not implemented")

class UniformUncertainty(Uncertainty):
    """An uncertainty for real-valued parameters following a uniform distribution."""

    def __init__(self, name, min_value, max_value, **kwargs):
        super(UniformUncertainty, self).__init__(name)
        self.min_value = float(min_value)
        self.max_value = float(max_value)

    def levels(self, nlevels):
        d = (self.max_value - self.min_value) / nlevels
        result = []

        for i in range(nlevels):
            result.append(self.min_value + random.uniform(i*d, (i+1)*d))

        return result

    def ppf(self, x):
        return self.min_value + x*(self.max_value - self.min_value)


class TriangularUncertainty(Uncertainty):
    """An uncertainty with a triangular distribution."""

    def __init__(self, name, min_value, max_value, mode_value, **kwargs):
        super(TriangularUncertainty, self).__init__(name)
        self.min_value = float(min_value)
        self.max_value = float(max_value)
        self.mode_value = float(mode_value)

        if not (self.min_value < self.max_value):
            raise ValueError('Min cannot be less than max.')
        if not ((self.min_value <= self.mode_value) and (self.mode_value <= self.max_value)):
            raise ValueError('Mode must be between min and max.')

        # Paramters used by scipy.stats
        self.scale = self.max_value - self.min_value
        self.c = (self.mode_value - self.min_value) / self.scale

    def levels(self, nlevels):
        ulevels = UniformUncertainty(self.name, 0.0, 1.0).levels(nlevels)
        return stats.triang.ppf(ulevels, c=self.c, loc=self.min_value, scale=self.scale)

    def ppf(self, x):
        return stats.triang.ppf(x, c=self.c, loc=self.min_value, scale=self.scale)


class PointUncertainty(Uncertainty):
    """An uncertainty distribution with all its probability mass at one point on the real line."""

    def __init__(self, name, value):
        super(PointUncertainty, self).__init__(name)
        self.value = value

    def levels(self, nlevels):
        return [self.value] * nlevels

    def ppf(self, x):
        return self.value


class NormalUncertainty(Uncertainty):
    """An uncertainty for real-valued parameters following a normal (Gaussian) distribution."""

    def __init__(self, name, mean, stdev, **kwargs):
        super(NormalUncertainty, self).__init__(name)
        self.mean = float(mean)
        self.stdev = float(stdev)

    def levels(self, nlevels):
        ulevels = UniformUncertainty(self.name, 0.0, 1.0).levels(nlevels)
        return stats.norm.ppf(ulevels, self.mean, self.stdev)

    def ppf(self, x):
        return stats.norm.ppf(x, self.mean, self.stdev)

class LogNormalUncertainty(Uncertainty):
    """An uncertainty for real-valued parameters following a log normal distribution."""

    def __init__(self, name, mu, sigma, **kwargs):
        super(LogNormalUncertainty, self).__init__(name)
        self.mu = float(mu)
        self.sigma = float(sigma)

    def levels(self, nlevels):
        ulevels = UniformUncertainty(self.name, 0.0, 1.0).levels(nlevels)
        return self.mu*stats.lognorm.ppf(ulevels, self.sigma)

    def ppf(self, x):
        return self.mu*stats.lognorm.ppf(x, self.sigma)

class IntegerUncertainty(Uncertainty):
    """An uncertainty for integer parameters that follows a uniform distribution."""

    def __init__(self, name, min_value, max_value, **kwargs):
        super(IntegerUncertainty, self).__init__(name)
        self.min_value = int(min_value)
        self.max_value = int(max_value)

    def levels(self, nlevels):
        ulevels = UniformUncertainty(self.name, self.min_value, self.max_value+0.9999).levels(nlevels)
        return [int(math.floor(x)) for x in ulevels]

    def ppf(self, x):
        return int(math.floor(self.min_value + x*(self.max_value + 0.9999 - self.min_value)))

class CategoricalUncertainty(Uncertainty):
    """An uncertainty for categorial parameters that follows a uniform distribution."""

    def __init__(self, name, categories, **kwargs):
        super(CategoricalUncertainty, self).__init__(name)
        self.categories = categories

    def levels(self, nlevels):
        ilevels = IntegerUncertainty(self.name, 0, len(self.categories)-1).levels(nlevels)
        return [self.categories[i] for i in ilevels]

    def ppf(self, x):
        return self.categories[int(math.floor(x*(len(self.categories)-0.0001)))]

class NamedObjectMap(object):

    def __init__(self, type):
        super(NamedObjectMap, self).__init__()
        self.type = type
        self._data = OrderedDict()

        if not issubclass(type, NamedObject):
            raise TypeError("type must be a NamedObject")

    def clear(self):
        self._data = OrderedDict()

    def __len__(self):
        return len(self._data)

    def __getitem__(self, key):
        if isinstance(key, int):
            for i, (k, v) in enumerate(self._data.items()):
                if i == key:
                    return v
            raise KeyError(key)
        else:
            return self._data[key]

    def __setitem__(self, key, value):
        if not isinstance(value, self.type):
            raise TypeError("can only add " + self.type.__name__ + " objects")

        if isinstance(key, int):
            self._data = OrderedDict([(value.name, value) if i == key else (k, v) for i, (k, v) in enumerate(self._data.items())])
        else:
            if value.name != key:
                raise ValueError("key does not match name of " + self.type.__name__)

            self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data.values())

    def __contains__(self, item):
        return item in self._data

    def extend(self, value):
        if hasattr(value, "__iter__"):
            for item in value:
                if not isinstance(item, self.type):
                    raise TypeError("can only add " + self.type.__name__ + " objects")

            for item in value:
                self._data[item.name] = item
        elif isinstance(value, self.type):
            self._data[value.name] = value
        else:
            raise TypeError("can only add " + str(self.type) + " objects")

    def __add__(self, value):
        self.extend(value)
        return self

    def __iadd__(self, value):
        self.extend(value)
        return self

    def keys(self):
        return self._data.keys()

    # def __getattr__(self, name):
    #     return getattr(self._data, name)

class ParameterMap(NamedObjectMap):

    def __init__(self):
        super(ParameterMap, self).__init__(Parameter)

class ResponseMap(NamedObjectMap):

    def __init__(self):
        super(ResponseMap, self).__init__(Response)

class LeverMap(NamedObjectMap):

    def __init__(self):
        super(LeverMap, self).__init__(Lever)

class UncertaintyMap(NamedObjectMap):

    def __init__(self):
        super(UncertaintyMap, self).__init__(Uncertainty)

class Model(object):

    def __init__(self, function):
        super(Model, self).__init__()
        self.function = function
        self._parameters = ParameterMap()
        self._responses = ResponseMap()
        self.constraints = []
        self._levers = LeverMap()
        self._uncertainties = UncertaintyMap()
        self.fixed_parameters = {}

    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters.extend(value)

    @property
    def responses(self):
        return self._responses

    @responses.setter
    def responses(self, value):
        self._responses.extend(value)

    @property
    def levers(self):
        return self._levers

    @levers.setter
    def levers(self, value):
        self._levers.extend(value)

    @property
    def uncertainties(self):
        return self._uncertainties

    @uncertainties.setter
    def uncertainties(self, value):
        self._uncertainties.extend(value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        pass

    def fix(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, dict):
                self.fixed_parameters.update(arg)
            else:
                raise RhodiumError("fix() only accepts keyword arguments or a dict")

        self.fixed_parameters.update(kwargs)

class DataSet(list):

    def __init__(self, data=[]):
        super(DataSet, self).__init__()

        if isinstance(data, str):
            self.load(data)
        else:
            for entry in data:
                self.append(entry)

    def append(self, sample):
        if not isinstance(sample, dict):
            raise TypeError("DataSet can only contain dict objects")

        super(DataSet, self).append(sample)

    def __str__(self):
        result = ""

        if len(self) == 0:
            result += "Empty (no feasible solutions)"

        for i in range(len(self)):
            result += "Index "
            result += str(i)
            result += ":\n"

            for key in self[i]:
                result += "    "
                result += str(key)
                result += ": "
                result += str(self[i][key])
                result += "\n"

        return result

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __getitem__(self, pos):
        if isinstance(pos, tuple):
            indices, keys = pos

            if isinstance(indices, slice):
                indices = list(range(*indices.indices(len(self))))
            elif isinstance(indices, int):
                indices = [indices]

            if not isinstance(keys, list) and not isinstance(keys, tuple):
                keys = [keys]

            result = DataSet()

            for i in indices:
                submap = {}

                for key in keys:
                    submap[key] = super(DataSet, self).__getitem__(i)[key]

                result.append(submap)

            return result
        elif isinstance(pos, str):
            return self.as_list(pos)
        elif isinstance(pos, slice):
            indices = list(range(*pos.indices(len(self))))
            result = DataSet()

            for i in indices:
                result.append(super(DataSet, self).__getitem__(i))

            return result
        else:
            return super(DataSet, self).__getitem__(pos)

    def __setitem__(self, pos, value):
        if isinstance(pos, str):
            if isinstance(value, (list, tuple)) and len(value) == len(self):
                for i, o in enumerate(self):
                    o[pos] = value[i]
            else:
                for o in self:
                    o[pos] = value
        else:
            return super(DataSet, self).__setitem__(pos)

    def _trim(self, value, index=None):
        if index is not None and isinstance(value, (list, tuple)):
            return value[index]
        else:
            return value

    def as_list(self, key=None, index=None):
        result = []

        if len(self) == 0:
            return result

        if key is None:
            if len(self[0].keys()) > 1:
                raise ValueError("Can not convert DataSet to list that contains more than one key")
            else:
                key = list(self[0].keys())[0]

        for i in range(len(self)):
            value = super(DataSet, self).__getitem__(i)[key]
            result.append(self._trim(value, index))

        return result

    def as_dataframe(self, keys=None, index=None, include_dtypes=None, exclude_dtypes=None):
        dict = OrderedDict()

        if keys is None:
            if len(self) == 0:
                raise ValueError("dataset is empty")
            else:
                keys = self[0].keys()

        if isinstance(keys, str):
            keys = [keys]

        for key in keys:
            dict[key] = [self._trim(d[key], index) for d in self]

        df = pd.DataFrame(dict)

        if include_dtypes is not None or exclude_dtypes is not None:
            df = df.select_dtypes(include_dtypes, exclude_dtypes)

        return df

    def as_array(self, keys=None, index=None):
        import numpy

        if len(self) == 0:
            return numpy.empty([0])

        if keys is None:
            keys = list(self[0].keys())

        if isinstance(keys, str):
            keys = [keys]

        if isinstance(keys, set):
            keys = list(keys)

        if len(keys) == 1:
            key = keys[0]
            result = numpy.empty([len(self)], dtype=numpy.dtype(type(self[0][key])))

            for i, env in enumerate(self):
                result[i] = self._trim(env[key], index)
        else:
            dt = {"names": keys, "formats": [numpy.dtype(type(self[0][key])) for key in keys]}
            result = numpy.empty([len(self)], dtype=dt)

            for i, env in enumerate(self):
                result[i] = tuple(self._trim(env[key], index) for key in keys)

        return result

    def find(self, expr, inverse=False):
        result = DataSet()

        for entry, cond in zip(self, self.apply(expr)):
            if cond:
                result.append(entry)

        return result

    def apply(self, expr, update=True):
        return _evaluate_all(expr, self, update)

    def find_min(self, key):
        index, _ = min(enumerate([d[key] for d in self]), key=operator.itemgetter(1))
        return self[index]

    def find_max(self, key):
        index, _ = max(enumerate([d[key] for d in self]), key=operator.itemgetter(1))
        return self[index]

    def save(self, file, format=None, **kwargs):
        save(self, file, format, **kwargs)

def save(data, file, format=None, **kwargs):
    if isinstance(data, DataSet):
        data = data.as_dataframe()

    if format is None:
        _, format = os.path.splitext(file)

        if len(format) > 0 and format[0] == ".":
            format = format[1:]

    if format == "xls" or format == "xlsx":
        if "index" not in kwargs:
            kwargs["index"] = False

        data.to_excel(file, **kwargs)
    elif format == "csv":
        if "index" not in kwargs:
            kwargs["index"] = False

        data.to_csv(file, **kwargs)
    elif format == "json":
        data.to_json(file, **kwargs)
    elif format == "pkl":
        data.to_pickle(file, **kwargs)
    else:
        raise ValueError("unsupported file format '%s'" % str(format))

class _FileModel(Model):

    def __init__(self):
        super(_FileModel, self).__init__(self._evaluate)

    def _evaluate(self, **kwargs):
        raise NotImplementedError("models loaded from files do not support evaluation")

def load(file, format=None, parameters=[], **kwargs):

    if format is None:
        _, format = os.path.splitext(file)

        if len(format) > 0 and format[0] == ".":
            format = format[1:]

    if format == "xls" or format == "xlsx":
        df = pd.read_excel(file, **kwargs)
    elif format == "csv":
        df = pd.read_csv(file, **kwargs)
    elif format == "json":
        df = pd.read_json(file, **kwargs)
    elif format == "pkl":
        df = pd.read_pickle(file, **kwargs)
    else:
        raise ValueError("unsupported file format '%s'" % str(format))

    names = list(df.columns.values)
    data = DataSet()

    if isinstance(parameters, str):
        parameters = [parameters]

    for i in range(df.shape[0]):
        entry = {}

        for j in range(df.shape[1]):
            entry[names[j]] = df.iloc[i, j]

        data.append(entry)

    model = _FileModel()
    model.parameters = [Parameter(names[j] if isinstance(j, int) else j) for j in parameters]
    model.responses = [Response(names[j]) for j in range(df.shape[1]) if j not in parameters and names[j] not in parameters]

    return (model, data)

def _overwrite_generator(samples, fixed_parameters):
    if inspect.isgenerator(samples) or (hasattr(samples, '__iter__') and not isinstance(samples, dict)):
        for sample in samples:
            result = sample.copy()
            result.update(fixed_parameters)
            yield result
    else:
        result = samples.copy()
        result.update(fixed_parameters)
        yield result

def overwrite(samples, fixed_parameters):
    if inspect.isgenerator(samples) or (hasattr(samples, '__iter__') and not isinstance(samples, dict)):
        return _overwrite_generator(samples, fixed_parameters)
    else:
        result = samples.copy()
        result.update(fixed_parameters)
        return result

def _update_generator(samples, fixed_parameters):
    if inspect.isgenerator(samples) or (hasattr(samples, '__iter__') and not isinstance(samples, dict)):
        for sample in samples:
            result = fixed_parameters.copy()
            result.update(sample)
            yield result
    else:
        result = fixed_parameters.copy()
        result.update(samples)
        yield result

def update(samples, fixed_parameters):
    if inspect.isgenerator(samples) or (hasattr(samples, '__iter__') and not isinstance(samples, dict)):
        return _update_generator(samples, fixed_parameters)
    else:
        result = fixed_parameters.copy()
        result.update(samples)
        return result

def populate_defaults(model, samples):
    if isinstance(samples, dict):
        samples = [samples]

    argspec = inspect.getargspec(model.function)
    default_values = {k: v for k, v in zip(argspec.args[-len(argspec.defaults):], argspec.defaults)}

    for sample in samples:
        for parameter in model.parameters:
            if parameter.name not in sample:
                if parameter.default_value is not None:
                    sample[parameter.name] = parameter.default_value
                else:
                    sample[parameter.name] = default_values[parameter.name]



import math
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import brentq as root
from rhodium import *
from rhodium.decorators import *

@RhodiumModel()
@Responses(Minimize("max_P"), Maximize("utility"), Maximize("inertia"), Maximize("reliability"))
@Constraints("reliability >= 0.95")
def lake_problem(
         pollution_limit = Real(0.0, 0.1, length=100),
         b = Uniform(0.1, 0.45, default_value=0.42),        # decay rate for P in lake (0.42 = irreversible)
         q = Uniform(2.0, 4.5, default_value=2.0),          # recycling exponent
         mean = 0.02,                                       # mean of natural inflows
         stdev = 0.001,                                     # standard deviation of natural inflows
         alpha = 0.4,                                       # utility from pollution
         delta = 0.98,                                      # future utility discount rate
         nsamples = 100):                                   # monte carlo sampling of natural inflows)
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(pollution_limit)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(pollution_limit)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)

        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)

    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    inertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return (max_P, utility, inertia, reliability)

# Generate 1000 SOWs using a LHS sampling of the uncertainties (b and q)
SOWs = sample_lhs(lake_problem, 1000)

# Apply a fixed pollution_limit policy to each SOW
policy = { "pollution_limit" : [0.01]*100 }
inputs = update(SOWs, policy)

# Evaluate each SOW
setup_cache(file="example.cache")
output = cache("decorators_output", lambda: evaluate(lake_problem, inputs))

# Classify each SOW into two groups: Reliable and Unreliable
output.apply("classification = 'Reliable' if reliability > 0.5 else 'Unreliable'")
print(output)

# Use PRIM to learn how the uncertainties (b and q) affect the classification
p = Prim(output, "classification", coi="Reliable", include=lake_problem.uncertainties.keys())
box = p.find_box()
box.show_tradeoff()
plt.show()


from rhodium import *
from rhodium.rbridge import RModel

# Provide the R file and function name
#   TIP: If unable to locate R, the path can be set with RCMD=r"C:\Program Files\R\R-3.2.1\bin\R.exe"
model = RModel("lake.R", "lake.eval")

# The parameter names must match the R arguments exactly
model.parameters = [Parameter("pollution_limit"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

# List all outputs from the R function, which should return these values either as
# an unnamed array in the given order (e.g., c(max_P, utility, ...)) or as a named
# list (e.g., list(max_P=..., utility=..., ...)).
model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

# Specify the levers
model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

# Optimize the model using Rhodium
output = optimize(model, "NSGAII", 10000)
print(output)



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
import numpy as np
from rhodium.model import Constraint, Model, Parameter, Response, \
    CategoricalUncertainty, IntegerUncertainty, PointUncertainty, \
    TriangularUncertainty, UniformUncertainty

class TestConstraint(unittest.TestCase):

    def testSimpleConstraint(self):
        c = Constraint("x < 1")
        self.assertTrue(c.is_feasible({"x": 0}))
        self.assertFalse(c.is_feasible({"x": 1}))

        self.assertEqual(0, c.distance({"x": 0}))
        self.assertNotEqual(0, c.distance({"x": 1}))

    def testComplexConstraint(self):
        c = Constraint("x < 1 and y > 1")
        self.assertTrue(c.is_feasible({"x": 0, "y": 2}))
        self.assertFalse(c.is_feasible({"x": 0, "y": 1}))
        self.assertFalse(c.is_feasible({"x": 1, "y": 1}))

        self.assertEqual(0, c.distance({"x": 0, "y": 2}))
        self.assertNotEqual(0, c.distance({"x": 0, "y": 1}))
        self.assertNotEqual(0, c.distance({"x": 1, "y": 1}))

class TestModelParameters(unittest.TestCase):

    def test(self):
        m = Model("foo")
        p = Parameter("x")

        m.parameters = [p]

        self.assertEqual(1, len(m.parameters))
        self.assertEqual(p, m.parameters[0])
        self.assertEqual(p, m.parameters["x"])
        self.assertTrue("x" in m.parameters)

    def testOrder(self):
        m = Model("foo")
        p1 = Parameter("x")
        p2 = Parameter("y")
        p3 = Parameter("z")

        m.parameters = [p1, p2, p3]

        self.assertEqual(3, len(m.parameters))
        self.assertEqual(p1, m.parameters[0])
        self.assertEqual(p2, m.parameters[1])
        self.assertEqual(p3, m.parameters[2])

    def testInvalidType(self):
        m = Model("foo")
        with self.assertRaises(TypeError):
            m.parameters = [Response("x")]

class TestUniformUncertainty(unittest.TestCase):

    def testLevels(self):
        uu = UniformUncertainty("x", 0.0, 1.0)

        levels = uu.levels(50)
        self.assertTrue(all(i >= 0.0 and i <= 1.0 for i in levels))

        levels = uu.levels(3)
        self.assertTrue(all(i >= 0.0 and i <= 1.0 for i in levels))

    def testPpf(self):
        uu = UniformUncertainty("x", 0.0, 1.0)

        self.assertEqual(0.0, uu.ppf(0.0))
        self.assertEqual(0.5, uu.ppf(0.5))
        self.assertEqual(1.0, uu.ppf(1.0))

class TestIntegerUncertainty(unittest.TestCase):

    def testLevels(self):
        iu = IntegerUncertainty("x", 0, 10)

        levels = iu.levels(50)
        self.assertTrue(all(i >= 0 and i <= 10 for i in levels))
        self.assertTrue(all(isinstance(i, int) for i in levels))

        levels = iu.levels(3)
        self.assertTrue(all(i >= 0 and i <= 10 for i in levels))
        self.assertTrue(all(isinstance(i, int) for i in levels))

    def testPpf(self):
        iu = IntegerUncertainty("x", 0, 10)

        self.assertEqual(0, iu.ppf(0.0))
        self.assertEqual(5, iu.ppf(0.5))
        self.assertEqual(10, iu.ppf(1.0))

class TestCategoricalUncertainty(unittest.TestCase):

    def testLevels(self):
        categories = ["a", "b", "c"]
        cu = CategoricalUncertainty("x", categories)

        levels = cu.levels(50)
        self.assertTrue(all(i in categories for i in levels))

        levels = cu.levels(3)
        self.assertTrue(all(i in categories for i in levels))

    def testPpf(self):
        categories = ["a", "b", "c"]
        cu = CategoricalUncertainty("x", categories)

        self.assertEqual("a", cu.ppf(0.0))
        self.assertEqual("b", cu.ppf(0.5))
        self.assertEqual("c", cu.ppf(1.0))

class TestTriangularUncertainty(unittest.TestCase):

    def test_ppf_0_1_symmetric(self):
        """Check the PPF on a symmetric triangular distribution on (0, 1)."""
        tu = TriangularUncertainty('x', 0, 1, 0.5)

        self.assertEqual(tu.ppf(0), 0)
        self.assertEqual(tu.ppf(1), 1)
        self.assertEqual(tu.ppf(0.5), 0.5)

    def test_ppf_0_1_skew(self):
        """Check the PPF on a skewed triangular distribution on (0, 1)."""
        tu = TriangularUncertainty('x', 0, 1, 0.25)

        self.assertEqual(tu.ppf(0), 0)
        self.assertEqual(tu.ppf(1), 1)
        self.assertEqual(tu.ppf(0.25), 0.25)

    def test_ppf_10_20_skew(self):
        """Check the PPF on a skewed triangular distribution on (10, 20)."""
        tu = TriangularUncertainty('x', 10, 20, 12.5)

        self.assertEqual(tu.ppf(0), 10)
        self.assertEqual(tu.ppf(1), 20)
        self.assertEqual(tu.ppf(0.25), 12.5)

    def test_out_of_order(self):
        """Check that an error is raised if the min, max and mode are not properly ordered."""
        with self.assertRaises(ValueError):
            TriangularUncertainty('x', 2, 1, 0)
        with self.assertRaises(ValueError):
            TriangularUncertainty('x', 0, 1, 2)
        with self.assertRaises(ValueError):
            TriangularUncertainty('x', 2, 0, 1)

class TestPointUncertainty(unittest.TestCase):
    """Unit tests for PointUncertainty"""

    def test_ppf(self):
        """Check that the ppf returns the point value for any quantile."""
        value = 1
        pu = PointUncertainty('x', value)
        for quantile in np.linspace(0, 1):
            self.assertEqual(pu.ppf(quantile), value)


if __name__ == '__main__':
    unittest.main()



import math
import numpy as np
from scipy.optimize import brentq as root
from rhodium import *
from j3 import J3

# Construct the lake problem
def lake_problem(pollution_limit,
         b = 0.42,        # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,         # recycling exponent
         mean = 0.02,     # mean of natural inflows
         stdev = 0.001,   # standard deviation of natural inflows
         alpha = 0.4,     # utility from pollution
         delta = 0.98,    # future utility discount rate
         nsamples = 100): # monte carlo sampling of natural inflows
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(pollution_limit)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(pollution_limit)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)

        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)

    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    inertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return (max_P, utility, inertia, reliability)

model = Model(lake_problem)

# Define all parameters to the model that we will be studying
model.parameters = [Parameter("pollution_limit"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

# Define the model outputs
model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

# Define any constraints (can reference any parameter or response by name)
#model.constraints = [Constraint("reliability >= 0.95")]

# Some parameters are levers that we control via our policy
model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

# Some parameters are exogeneous uncertainties, and we want to better
# understand how these uncertainties impact our model and decision making
# process
model.uncertainties = [UniformUncertainty("b", 0.1, 0.45),
                       UniformUncertainty("q", 2.0, 4.5),
                       UniformUncertainty("mean", 0.01, 0.05),
                       UniformUncertainty("stdev", 0.001, 0.005),
                       UniformUncertainty("delta", 0.93, 0.99)]

# Prepare the cache for storing intermediate results
setup_cache(file="example.cache")

# Optimize the model or get cached results if they exist.  Note that the
# call to optimize is wrapped in a lambda function to enable lazy evaluation.
output = cache("output", lambda: optimize(model, "NSGAII", 10000))

# Visualize using J3
J3(output.as_dataframe(list(model.responses.keys())))


# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from ..model import Parameter, Response
from ..optimization import evaluate
from ..openmdao import OpenMDAOModel

class TestOpenMDAOModel(unittest.TestCase):

    def test(self):
        try:
            from openmdao.api import IndepVarComp, Component, Problem, Group
        except ImportError as e:
            self.skipTest(f"OpenMDAO not available: {e}")

        # the following example is taken from OpenMDAO's Paraboloid tutorial
        class Paraboloid(Component):
            """ Evaluates the equation f(x,y) = (x-3)^2 + xy + (y+4)^2 - 3 """

            def __init__(self):
                super(Paraboloid, self).__init__()

                self.add_param('x', val=0.0)
                self.add_param('y', val=0.0)

                self.add_output('f_xy', val=0.0)

            def solve_nonlinear(self, params, unknowns, resids):
                """f(x,y) = (x-3)^2 + xy + (y+4)^2 - 3
                Optimal solution (minimum): x = 6.6667; y = -7.3333
                """

                x = params['x']
                y = params['y']

                unknowns['f_xy'] = (x-3.0)**2 + x*y + (y+4.0)**2 - 3.0

            def linearize(self, params, unknowns, resids):
                """ Jacobian for our paraboloid."""

                x = params['x']
                y = params['y']
                J = {}

                J['f_xy', 'x'] = 2.0*x - 6.0 + y
                J['f_xy', 'y'] = 2.0*y + 8.0 + x
                return J

        top = Problem()
        root = top.root = Group()

        root.add('p1', IndepVarComp('x', 3.0))
        root.add('p2', IndepVarComp('y', -4.0))
        root.add('p', Paraboloid())

        root.connect('p1.x', 'p.x')
        root.connect('p2.y', 'p.y')

        top.setup()

        # here is where Rhodium begins
        model = OpenMDAOModel(top)
        model.parameters = [Parameter("x", connect="p1.x"),
                            Parameter("y", connect="p2.y")]
        model.responses = [Response("f_xy", connect="p.f_xy")]

        output = evaluate(model, {"x": 3.0, "y": -4.0})
        self.assertEqual(-15.0, output["f_xy"])



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import os
import platform
import subprocess
import unittest
from ..model import Parameter, Response
from ..optimization import evaluate
from ..ffi import NativeModel

class TestNativeModel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # determine the relative path to the source file
        dir = os.path.dirname(__file__)
        cwd = os.getcwd()
        common_prefix = os.path.commonprefix([dir, cwd])
        rel_dir = os.path.relpath(dir, cwd)
        src = os.path.join(rel_dir, "test.c")

        if platform.system() == "Linux":
            libname = "libtest.so"
        elif platform.system() == "Darwin":
            libname = "libtest.dylib"
        elif platform.system() == "Windows":
            libname = "test.dll"

        result = subprocess.run(["gcc", "-shared", "-fPIC", "-o", libname, src], capture_output=True)
        print(result.stdout)
        print(result.stderr)
        result.check_returncode()

        cls.sopath = os.path.join(os.getcwd(), libname)

    def testNormalReturn(self):
        model = NativeModel(TestNativeModel.sopath, "norm_return")
        model.parameters = [Parameter("x", type="double"),
                            Parameter("y", type="double")]
        model.responses = [Response("z", type="double")]
        result = evaluate(model, {"x": 3, "y": 5})
        self.assertEqual(15, result["z"])

    def testArgumentReturn(self):
        model = NativeModel(TestNativeModel.sopath, "arg_return")
        model.parameters = [Parameter("x", type="double"),
                            Parameter("y", type="double")]
        model.responses = [Response("z", type="double", asarg=True)]
        result = evaluate(model, {"x": 3, "y": 5})
        self.assertEqual(15, result["z"])

    def testSum(self):
        model = NativeModel(TestNativeModel.sopath, "sum")
        model.parameters = [Parameter("x", type="double*10")]
        model.responses = [Response("sum", type="double")]
        result = evaluate(model, {"x": [1, 2, 3, 4, 5]})
        self.assertEqual(15, result["sum"])

    def testArrayAdd(self):
        model = NativeModel(TestNativeModel.sopath, "array_add")
        model.parameters = [Parameter("x", type="double*", len_arg="n"),
                            Parameter("y", type="double*", len_arg="n"),
                            Parameter("n", type="int")]
        model.responses = [Response("z", type="double*", len_arg="n", asarg=True)]
        result = evaluate(model, {"x": [1, 2, 3, 4, 5], "y": [5, 4, 3, 2, 1]})
        self.assertEqual([6, 6, 6, 6, 6], result["z"])



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
from rhodium import Model, Parameter, Response, IntegerLever, RealLever, \
    optimize

def schaffer(x):
    return [x**2, (x-2)**2]

class TestOptimization(unittest.TestCase):

    def testReal(self):
        model = Model(schaffer)

        model.parameters = [Parameter("x")]
        model.responses = [Response("f1", Response.MINIMIZE),
                           Response("f2", Response.MINIMIZE)]
        model.levers = [RealLever("x", -10.0, 10.0)]

        output = optimize(model, "NSGAII", 10000)

        self.assertTrue(len(output) > 0)

        for i in range(len(output)):
            self.assertTrue("x" in output[i])
            self.assertTrue("f1" in output[i])
            self.assertTrue("f2" in output[i])

    def testInteger(self):
        model = Model(schaffer)

        model.parameters = [Parameter("x")]
        model.responses = [Response("f1", Response.MINIMIZE),
                           Response("f2", Response.MINIMIZE)]
        model.levers = [IntegerLever("x", -10, 10)]

        output = optimize(model, "NSGAII", 10000)

        self.assertTrue(len(output) > 0)

        for i in range(len(output)):
            self.assertTrue("x" in output[i])
            self.assertTrue(isinstance(output[i]["x"], int))
            self.assertTrue("f1" in output[i])
            self.assertTrue("f2" in output[i])



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import sys
import shelve
import atexit
import pickle
import inspect
import warnings
import functools

_CACHE_FILE = None
_CACHE_CLEAR = False
_CACHE_WARNINGS = True
_CACHE = None

def setup_cache(file=None, clear=False, warn=True):
    global _CACHE_FILE
    global _CACHE_CLEAR
    global _CACHE_WARNINGS
    global _CACHE

    if _CACHE:
        if warn:
            warnings.warn("cache already opened", UserWarning)
        return

    _CACHE_FILE = file
    _CACHE_CLEAR = clear
    _CACHE_WARNINGS = warn

def _do_open():
    global _CACHE_FILE
    global _CACHE_CLEAR
    global _CACHE

    if _CACHE is not None:
        return

    if not _CACHE_FILE:
        _CACHE_FILE = "rhodium.cache"

    _CACHE = shelve.open(_CACHE_FILE, flag='n' if _CACHE_CLEAR else 'c')

    atexit.register(lambda: _do_close())

def _do_close():
    if _CACHE is not None:
        _CACHE.close()

def cached(func):

    @functools.wraps(func)
    def inner(*args, **kwargs):
        try:
            arguments = (args, kwargs)

            if sys.version_info[0] == 2:
                arguments_str = pickle.dumps(arguments)
            else:
                arguments_str = pickle.dumps(arguments, protocol=0).decode('ascii')

            return cache(func.__name__ + arguments_str, lambda: func(*args, **kwargs))
        except pickle.PickleError:
            if _CACHE_WARNINGS:
                warnings.warn("arguments not pickleable, unable to cache value", UserWarning)
            return func(*args, **kwargs)

    return inner

def clear_cache(*args):
    _do_open()

    for arg in args:
        if arg in _CACHE:
            del _CACHE[arg]

def cache(key, value):
    """Simple file cache using shelve.

    Aims to be a very simple wrapper for shelve.  Suppose we are performing
    an expensive function evaluation we want to cache:

        data = compute(args)

    we can cache the result and load it if available:

        data = cache("data", lambda: compute(args))

    Note the use of lambda to enable lazy evaluation of the function.
    """
    _do_open()

    if not inspect.isfunction(value):
        if _CACHE_WARNINGS:
            warnings.warn("cache value should be a function to enable lazy evaluation", UserWarning)

    if key not in _CACHE:
        if inspect.isfunction(value):
            _CACHE[key] = value()
        else:
            _CACHE[key] = value

    return _CACHE[key]



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import unittest
import pandas as pd
from ..brush import Brush, BrushSet, apply_brush
from ..config import RhodiumConfig

class TestBrush(unittest.TestCase):

    def testSimpleBrush(self):
        d = {'col1': [1, 2], 'col2': [3, 4]}
        df = pd.DataFrame(data=d)
        brush = Brush("col1 < 2")
        assignment = apply_brush(brush, df)
        self.assertEqual(assignment[0], brush.name)
        self.assertEqual(assignment[1], RhodiumConfig.default_unassigned_label)

    def testComplexBrush(self):
        d = {'col1': [1, 2, 3], 'col2': [3, 4, 5]}
        df = pd.DataFrame(data=d)
        brush = Brush("col1 <= 2 and col2 >= 4")
        assignment = apply_brush(brush, df)
        self.assertEqual(assignment[0], RhodiumConfig.default_unassigned_label)
        self.assertEqual(assignment[1], brush.name)
        self.assertEqual(assignment[2], RhodiumConfig.default_unassigned_label)

    def testMultipleBrush(self):
        d = {'col1': [1, 2, 3], 'col2': [3, 4, 5]}
        df = pd.DataFrame(data=d)
        brush1 = Brush("col1 < 2")
        brush2 = Brush("col2 > 4")
        assignment = apply_brush(BrushSet([brush1, brush2]), df)
        self.assertEqual(assignment[0], brush1.name)
        self.assertEqual(assignment[1], RhodiumConfig.default_unassigned_label)
        self.assertEqual(assignment[2], brush2.name)



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from platypus.config import PlatypusConfig

class _RhodiumConfig(object):

    def __init__(self):
        super(_RhodiumConfig, self).__init__()
        self.default_cmap = plt.get_cmap("rainbow")
        self.default_brush_colors = sns.color_palette()
        self.default_unassigned_brush_color = mpl.colors.colorConverter.to_rgb("#CCCCCC")
        self.default_unassigned_label = "Unassigned"

    @property
    def default_evaluator(self):
        return PlatypusConfig.default_evaluator

    @default_evaluator.setter
    def default_evaluator(self, value):
        PlatypusConfig.default_evaluator = value

    @property
    def default_log_frequency(self):
        return PlatypusConfig.default_log_frequency

    @default_log_frequency.setter
    def default_log_frequency(self, value):
        PlatypusConfig.default_log_frequency = value

RhodiumConfig = _RhodiumConfig()



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
from pyper import R
from .model import Model
import numpy as np

class RModel(Model):

    def __init__(self, file, function, **kwargs):
        super(RModel, self).__init__(self._evaluate)
        self.file = file
        self.r_function = function
        self.r = R(**kwargs)

        with open(file) as f:
            self.r(f.read())

    def _evaluate(self, **kwargs):
        prefix = "...rhodium."
        assigned_parameters = []

        for parameter in self.parameters:
            if parameter.name in kwargs:
                self.r.assign(prefix + parameter.name, kwargs[parameter.name])
                assigned_parameters.append(parameter)

        self.r(prefix + "..result = " + self.r_function + "(" + ",".join([prefix + p.name for p in assigned_parameters]) + ")")
        r_result = self.r[prefix + "..result"]

        result = {}

        if isinstance(r_result, (list, tuple, np.ndarray)):
            for i, response in enumerate(self.responses):
                result[response.name] = r_result[i]
        elif isinstance(r_result, dict):
            for response in self.responses:
                result[response.name] = r_result[response.name]
        else:
            if len(self.responses) > 1:
                raise ValueError("received more than one response from R")
            else:
                result[self.responses[0].name] = r_result

        return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
__author__ = "David Hadka"
__copyright__ = "Copyright 2015-2024, David Hadka"
__license__ = "GPLv3"
__version__ = "1.4.0"
__maintainer__ = "David Hadka"
__email__ = "dhadka@users.noreply.github.com"

from .config import *
from .model import *
from .plot import *
from .sa import *
from .classification import *
from .cache import *
from .optimization import *
from .sampling import *
from .robustness import *
from .brush import *
from .utils import *

from platypus.evaluator import *



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import inspect
import functools
from collections import OrderedDict
from platypus import Job, Problem, unique, nondominated
from .model import Response, DataSet

def generate_jobs(model, samples):
    if isinstance(samples, dict):
        yield EvaluateJob(model, samples)
    else:
        for sample in samples:
            yield EvaluateJob(model, sample)

class EvaluateJob(Job):

    def __init__(self, model, sample):
        super(EvaluateJob, self).__init__()
        self.model = model
        self.sample = sample
        self._args = inspect.getfullargspec(model.function).args

    def run(self):
        # populate model arguments
        args = {}

        for parameter in self.model.parameters:
            if parameter.name in self.sample:
                args[parameter.name] = self.sample[parameter.name]
            elif parameter.default_value is not None:
                args[parameter.name] = parameter.default_value

        # evaluate the model
        raw_output = self.model.function(**args)

        # support output as a dict or list-like object
        if isinstance(raw_output, dict):
            args.update(raw_output)
        elif isinstance(raw_output, (list, tuple)):
            offset = 0

            for response in self.model.responses:
                if response.dir is not Response.IGNORE:
                    args[response.name] = raw_output[offset]
                    offset += 1
        elif len(self.model.responses) == 1 and self.model.responses[0].dir != Response.IGNORE:
            args[self.model.responses[0].name] = raw_output

        self.output = args

def evaluate(model, samples, evaluator=None, log_frequency=None):
    if evaluator is None:
        from .config import RhodiumConfig
        evaluator = RhodiumConfig.default_evaluator

    if log_frequency is None:
        from .config import RhodiumConfig
        log_frequency = RhodiumConfig.default_log_frequency

    if (inspect.isgenerator(samples) or (hasattr(samples, '__iter__')) and not isinstance(samples, dict)):
        results = evaluator.evaluate_all(generate_jobs(model, samples), job_name="Evaluate", log_frequency=log_frequency)
        return DataSet([result.output for result in results])
    else:
        results = evaluator.evaluate_all(generate_jobs(model, samples), job_name="Evaluate", log_frequency=log_frequency)
        return results[0].output

def _evaluation_function(vars, model, nvars, nobjs, nconstrs, levers):
    env = {}
    offset = 0

    for lever, length in levers:
        env[lever.name] = lever.from_variables(vars[offset:(offset+length)])
        offset += length

    job = EvaluateJob(model, env)
    job.run()

    objectives = [job.output[r.name] for r in model.responses if r.dir in [Response.MINIMIZE, Response.MAXIMIZE]]
    constraints = [constraint.distance(job.output) for constraint in model.constraints]

    if nconstrs > 0:
        return objectives, constraints
    else:
        return objectives

def _to_problem(model):
    variables = []
    levers = []

    for lever in model.levers:
        vars = lever.to_variables()
        variables.extend(vars)
        levers.append((lever, len(vars)))

    nvars = len(variables)
    nobjs = sum([1 if r.dir == Response.MINIMIZE or r.dir == Response.MAXIMIZE else 0 for r in model.responses])
    nconstrs = len(model.constraints)

    function = functools.partial(_evaluation_function,
                                 model=model,
                                 nvars=nvars,
                                 nobjs=nobjs,
                                 nconstrs=nconstrs,
                                 levers=levers)

    problem = Problem(nvars, nobjs, nconstrs, function)
    problem.types[:] = variables
    problem.directions[:] = [Problem.MINIMIZE if r.dir == Response.MINIMIZE else Problem.MAXIMIZE for r in model.responses if r.dir == Response.MINIMIZE or r.dir == Response.MAXIMIZE]
    problem.constraints[:] = "==0"
    return (problem, levers)

def optimize(model, algorithm="NSGAII", NFE=10000, module="platypus", **kwargs):
    if not model.levers:
        raise AssertionError("must have levers to optimize")

    module = __import__(module, fromlist=[''])
    class_ref = getattr(module, algorithm)

    args = kwargs.copy()
    args["problem"], levers = _to_problem(model)

    instance = class_ref(**args)
    instance.run(NFE)

    result = DataSet()

    for solution in unique(nondominated(instance.result)):
        if not solution.feasible:
            continue

        env = OrderedDict()
        offset = 0

        # decode from Platypus' internal representation (this should be fixed in Platypus instead)
        vars = [solution.problem.types[i].decode(solution.variables[i]) for i in range(solution.problem.nvars)]

        for lever, length in levers:
            env[lever.name] = lever.from_variables(vars[offset:(offset+length)])
            offset += length

        if any([r.dir not in [Response.MINIMIZE, Response.MAXIMIZE] for r in model.responses]):
            # if there are any responses not included in the optimization, we must
            # re-evaluate the model to get all responses
            env = evaluate(model, env)
        else:
            for i, response in enumerate(model.responses):
                env[response.name] = solution.objectives[i]

        result.append(env)

    return result

def _robust_evaluation_function(vars, model, SOWs, nvars, nobjs, nconstrs, levers, obj_aggregate, constr_aggregate):
    objectives = {}
    constraints = {}

    for response in model.responses:
        if response.dir in [Response.MINIMIZE, Response.MAXIMIZE]:
            objectives[response] = []

    for constraint in model.constraints:
        constraints[constraint] = []

    for SOW in SOWs:
        env = SOW.copy()
        offset = 0

        for lever, length in levers:
            env[lever.name] = lever.from_variables(vars[offset:(offset+length)])
            offset += length

        job = EvaluateJob(model, env)
        job.run()

        for response in model.responses:
            if response.dir in [Response.MINIMIZE, Response.MAXIMIZE]:
                objectives[response].append(job.output[response.name])

        for constraint in model.constraints:
            constraints[constraint].append(constraint.distance(job.output))

    if isinstance(obj_aggregate, dict):
        objective_values = [obj_aggregate[r.name](objectives[r]) for r in model.responses if r.dir in [Response.MINIMIZE, Response.MAXIMIZE]]
    else:
        objective_values = [obj_aggregate(objectives[r]) for r in model.responses if r.dir in [Response.MINIMIZE, Response.MAXIMIZE]]

    if isinstance(constr_aggregate, dict):
        constraint_values = [constr_aggregate[c](constraints[c]) for c in model.constraints]
    else:
        constraint_values = [constr_aggregate(constraints[c]) for c in model.constraints]

    if nconstrs > 0:
        return objective_values, constraint_values
    else:
        return objective_values

def _to_robust_problem(model, SOWs, obj_aggregate, constr_aggregate):
    variables = []
    levers = []

    for lever in model.levers:
        vars = lever.to_variables()
        variables.extend(vars)
        levers.append((lever, len(vars)))

    nvars = len(variables)
    nobjs = sum([1 if r.dir == Response.MINIMIZE or r.dir == Response.MAXIMIZE else 0 for r in model.responses])
    nconstrs = len(model.constraints)

    function = functools.partial(_robust_evaluation_function,
                                 model=model,
                                 SOWs=SOWs,
                                 nvars=nvars,
                                 nobjs=nobjs,
                                 nconstrs=nconstrs,
                                 levers=levers,
                                 obj_aggregate=obj_aggregate,
                                 constr_aggregate=constr_aggregate)

    problem = Problem(nvars, nobjs, nconstrs, function)
    problem.types[:] = variables
    problem.directions[:] = [Problem.MINIMIZE if r.dir == Response.MINIMIZE else Problem.MAXIMIZE for r in model.responses if r.dir == Response.MINIMIZE or r.dir == Response.MAXIMIZE]
    problem.constraints[:] = "==0"
    return (problem, levers)

def robust_optimize(model, SOWs, algorithm="NSGAII", NFE=10000, obj_aggregate=None, constr_aggregate=None, **kwargs):
    module = __import__("platypus", fromlist=[''])
    class_ref = getattr(module, algorithm)

    if obj_aggregate is None:
        from .robustness import mean
        obj_aggregate = mean

    if constr_aggregate is None:
        constr_aggregate = max

    args = kwargs.copy()
    args["problem"], levers = _to_robust_problem(model, SOWs, obj_aggregate, constr_aggregate)

    instance = class_ref(**args)
    instance.run(NFE)

    result = DataSet()

    for solution in unique(nondominated(instance.result)):
        if not solution.feasible:
            continue

        env = OrderedDict()
        offset = 0

        # decode from Platypus' internal representation (this should be fixed in Platypus instead)
        vars = [solution.problem.types[i].decode(solution.variables[i]) for i in range(solution.problem.nvars)]

        for lever, length in levers:
            env[lever.name] = lever.from_variables(vars[offset:(offset+length)])
            offset += length

        if any([r.dir not in [Response.MINIMIZE, Response.MAXIMIZE] for r in model.responses]):
            # if there are any responses not included in the optimization, we must
            # re-evaluate the model to get all responses
            env = evaluate(model, env)

        # here we copy over the objectives from the evaluated solution, which has been aggregated over all SOWs
        for i, response in enumerate([r for r in model.responses if r.dir in [Response.MINIMIZE, Response.MAXIMIZE]]):
            env[response.name] = solution.objectives[i]

        result.append(env)

    return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import pydot
import operator
import functools
import numpy as np
import numpy.lib.recfunctions as rf
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import sklearn
from sklearn import tree
from packaging.version import Version
from prim import Prim
from io import BytesIO, StringIO

class Cart(object):

    def __init__(self,
                 x,
                 y,
                 threshold=None,
                 threshold_type=">",
                 include=None,
                 exclude=None,
                 **kwargs):
        """Generates a decision tree for classification.

        Parameters
        ----------
        x : a matrix-like object (pandas.DataFrame, numpy.recarray, etc.)
            the independent variables
        y : a list-like object, the column name (str), or callable
            the dependent variable either provided as a list-like object
            classifying the data into cases of interest (e.g., False/True),
            a list-like object storing the raw variable value (in which case
            a threshold must be given), a string identifying the dependent
            variable in x, or a function called on each row of x to compute the
            dependent variable
        threshold : float
            threshold for identifying cases of interest
        threshold_type : str
            comparison operator used when identifying cases of interest
        include : list of str
            the names of variables included in the PRIM analysis
        exclude : list of str
            the names of variables excluded from the PRIM analysis
        """
        super(Cart, self).__init__()

        # Ensure the input x is a numpy matrix/array
        if isinstance(x, pd.DataFrame):
            x = x.to_records(index=False)
        elif isinstance(x, np.ma.MaskedArray):
            pass
        else:
            x = pd.DataFrame(x).to_records(index=False)

        # if y is a string or function, compute the actual response value
        # otherwise, ensure y is a numpy matrix/array
        if isinstance(y, str):
            key = y
            y = x[key]

            if exclude:
                exclude = list(exclude) + [key]
            else:
                exclude = [key]
        elif callable(y):
            fun = y
            y = np.apply_along_axis(fun, 0, x)
        elif isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y = y.values
        elif isinstance(y, np.ma.MaskedArray):
            pass
        else:
            y = np.asarray(y)

        # convert include/exclude arguments to lists if they are strings
        if include and isinstance(include, str):
            include = [include]

        if exclude and isinstance(exclude, str):
            exclude = [exclude]

        # include or exclude columns from the analysis
        if include:
            if isinstance(include, str):
                include = [include]

            drop_names = set(rf.get_names(x.dtype))-set(include)
            x = rf.drop_fields(x, drop_names, asrecarray=True)

        if exclude:
            if isinstance(exclude, str):
                exclude = [exclude]

            drop_names = set(exclude)
            x = rf.drop_fields(x, drop_names, asrecarray=True)

        # apply the threshold if
        if threshold:
            if callable(threshold):
                y = np.apply_along_axis(threshold, 0, y)
            else:
                # The syntax for threshold_type is "x <op> <threshold>", e.g.,
                # "x > 0.5".  However, partial only supports positional
                # arguments for built-in operators.  Thus, we must assign the
                # threshold to the first position and use a different operator.
                # For example, "x > 0.5" must be evaluated as "0.5 < x".
                OPERATORS = {"<": operator.ge,
                             ">": operator.le,
                             "<=": operator.gt,
                             ">=": operator.lt,
                             "=": operator.eq}

                op = OPERATORS[threshold_type]
                y = np.apply_along_axis(functools.partial(op, threshold), 0, y)

        # validate inputs
        if len(y.shape) > 1:
            raise ValueError("y is not a 1-d array")

        # extract feature names
        feature_names = rf.get_names(x.dtype)

        # ensure x is formatted as a 2D matrix
        x = x.view("<f8").reshape(x.shape + (-1,))

        clf = tree.DecisionTreeClassifier(**kwargs)
        clf = clf.fit(x, y)

        # add our custom metadata to the classifier
        self._feature_names = feature_names
        self._x = x
        self._y = y
        self._clf = clf

    def _get_names(self, **kwargs):
        clf = self._clf
        feature_names = kwargs.get("feature_names", None)
        class_names = kwargs.get("class_names", None)

        if feature_names is None:
            feature_names = self._feature_names

        if class_names is None:
            class_names = clf.classes_

        return feature_names, class_names

    def _create_graph(self, **kwargs):
        clf = self._clf
        dot_data = StringIO()
        feature_names, class_names = self._get_names(**kwargs)

        if Version(sklearn.__version__) >= Version('0.17'):
            tree.export_graphviz(clf,
                                 out_file=dot_data,
                                 feature_names=feature_names,
                                 class_names=class_names,
                                 filled=kwargs.get("filled", True),
                                 rounded=kwargs.get("rounded", True),
                                 special_characters=kwargs.get("special_characters", True),
                                 **kwargs)
        else:
            tree.export_graphviz(clf,
                                 out_file=dot_data,
                                 feature_names=feature_names,
                                 **kwargs)

        return pydot.graph_from_dot_data(dot_data.getvalue())[0]

    def __str__(self):
        return self._to_string()

    def print_tree(self, coi=None, all=True, **kwargs):
        results, node_list = self._to_string(coi, all, **kwargs)
        print(results)
        return node_list

    def _to_string(self, coi=None, all=True, **kwargs):
        result = ""
        clf = self._clf
        feature_names, class_names = self._get_names(**kwargs)

        if not hasattr(coi, "__iter__") and not isinstance(coi, str):
            coi = [coi]

        left = clf.tree_.children_left
        right = clf.tree_.children_right
        threshold = clf.tree_.threshold
        features = [feature_names[i] for i in clf.tree_.feature]
        classes = [class_names[np.argmax(i)] for i in clf.tree_.value]

        # get ids of the nodes to print
        if all:
            idx = range(1, clf.tree_.node_count)
        else:
            idx = np.argwhere(left == -1)[:, 0]

        def recurse(left, right, child, lineage=None):
            if lineage is None:
                lineage = []
            if child in left:
                parent = np.where(left == child)[0].item()
                split = "l"
            elif child in right:
                parent = np.where(right == child)[0].item()
                split = "r"

            lineage.append((features[parent], "<=" if split == "l" else ">", threshold[parent]))

            if parent == 0:
                lineage.reverse()
                return lineage
            else:
                return recurse(left, right, parent, lineage)

        # NOTE: This is where I added a child_list, where each scenario node should be added. Each child/scenario node is a dict. Probably, there exists more elegant solutions.
        child_list = []
        for child in idx:
            child_dict = {}
            if coi is None or classes[child] in coi:
                if len(result) > 0:
                    result += "\n"

                result += "Node %d: %s\n" % (child, classes[child])

                if coi is not None:
                    value = clf.tree_.value[child][0]
                    ncoi = sum([value[i] if class_names[i] in coi else 0 for i in range(clf.n_classes_)])
                    density = ncoi/np.sum(value)
                    coverage = ncoi/sum([1 if yi in coi else 0 for yi in self._y])
                    result += "    Density: %.2f%%\n" % (100*density,)
                    result += "    Coverage: %.2f%%\n" % (100*coverage,)
                # Below is original code (commented out). I split it up into two lines, to be able to save the rules of each child/scenario node in a separate vector.
                # result += "    Rule: " + " and\n          ".join(self._collapse_bounds(recurse(left, right, child), feature_names))
                rules_vec = self._collapse_bounds(recurse(left, right, child), feature_names)
                child_dict["Rules"] = rules_vec
                result += "    Rule: " + " and\n          ".join(rules_vec)

                # Now it is easy to just store the rest of the scenario data, and append the scenario node.
                child_dict["Node"] = str(child)
                child_dict["Class"] = str(classes[child])
                child_dict["Density"] = str(100*density)
                child_dict["Coverage"] = str(100*coverage)
                child_list.append(child_dict)    #

        # Now the code returns a list of the scenario nodes, which was what we wanted! But since I do not understand this code completely (e.g. isinstance(), the recursive calls,
        # or _collapse_bounds()) it is tricky to write up a smarter/more elegant solution.
        return result, child_list

    def _collapse_bounds(self, rules, keys):
        bounds = {}

        for rule in rules:
            if rule[0] in bounds:
                if rule[1] == "<=" and rule[2] <= bounds[rule[0]][1]:
                    bounds[rule[0]][1] = rule[2]
                elif rule[1] == ">" and rule[2] > bounds[rule[0]][0]:
                    bounds[rule[0]][0] = rule[2]
            else:
                bounds[rule[0]] = [rule[2] if rule[1] == ">" else -np.inf,
                                   rule[2] if rule[1] == "<=" else np.inf]

        result = []

        for key in keys:
            if key in bounds:
                if np.isinf(bounds[key][0]):
                    rule = "%s <= %f" % (key, bounds[key][1])
                elif np.isinf(bounds[key][1]):
                    rule = "%s > %f" % (key, bounds[key][0])
                else:
                    rule = "%f <= %s <= %f" % (bounds[key][0], key, bounds[key][1])

                result.append(rule)

        return result

    def show_tree(self, **kwargs):
        graph = self._create_graph(**kwargs)

        if "inline" in mpl.get_backend():
            # running inline in IPython
            from IPython.display import Image
            return Image(graph.create_png())
        else:
            # otherwise show within matplotlib
            img_data = BytesIO(graph.create_png())
            img = mpimg.imread(img_data)
            fig = plt.imshow(img)
            fig.axes.get_xaxis().set_visible(False)
            fig.axes.get_yaxis().set_visible(False)
            return fig

    def save(self, file, format="png", **kwargs):
        graph = self._create_graph(**kwargs)
        graph.write(file, format=format)

    def save_pdf(self, file, feature_names=None, **kwargs):
        self.save(file, "pdf", feature_names, **kwargs)

    def save_png(self, file, feature_names=None, **kwargs):
        self.save(file, "png", feature_names, **kwargs)

    def __getattr__(self, name):
        return getattr(self._clf, name)



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import random
from collections import OrderedDict
from .model import DataSet

def sample_uniform(model, nsamples):
    """Returns a data set with uncertainty parameters sampled using a uniform distribution."""
    result = DataSet()

    for i in range(nsamples):
        entry = OrderedDict()

        for uncertainty in model.uncertainties:
            entry[uncertainty.name] = uncertainty.ppf(random.uniform(0.0, 1.0))

        result.append(entry)

    return result

def sample_lhs(model, nsamples):
    """Returns a data set with uncertainty parameters sampled using Latin hypercube sampling."""
    if len(model.uncertainties) == 0:
        raise ValueError("model has no uncertainties defined")

    samples = OrderedDict()

    for uncertainty in model.uncertainties:
        levels = uncertainty.levels(nsamples)
        random.shuffle(levels)
        samples[uncertainty.name] = levels

    result = DataSet()

    for i in range(nsamples):
        entry = OrderedDict()

        for key, values in samples.items():
            entry[key] = values[i]

        # include model fixed parameters
        for key in model.fixed_parameters.keys():
            entry[key] = model.fixed_parameters[key]

        result.append(entry)

    return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np
import scipy.spatial as sp
from .model import DataSet, Response, populate_defaults, update
from .sampling import sample_lhs
from .optimization import evaluate

def _is_feasible(model, result):
    for constraint in model.constraints:
        if hasattr(constraint, "__call__"):
            if not constraint(result.copy()):
                return False
        else:
            if not eval(constraint.expr, result.copy()):
                return False

    return True

def check_feasibility(model, results):
    if isinstance(results, dict):
        return _is_feasible(model, results)
    else:
        return [_is_feasible(model, result) for result in results]

def mean(values):
    return float(sum(values)) / len(values)

def regret_type1(model, results, baseline, percentile=90):
    quantiles = []

    for response in model.responses:
        if response.dir == Response.MINIMIZE or response.dir == Response.MAXIMIZE:
            values = [abs((result[response.name] - baseline[response.name]) / baseline[response.name]) for result in results]
            quantiles.append(np.percentile(values, percentile))

    return max(quantiles)

def regret_type2(model, all_results, baseline_results, percentile=90):
    # for each uncertainty sampling, find the best value
    best = []

    for i in range(len(all_results[0])):
        entry = {}

        for response in model.responses:
            if response.dir == Response.MINIMIZE:
                entry[response.name] = min([result[i][response.name] for result in all_results])
            elif response.dir == Response.MAXIMIZE:
                entry[response.name] = max([result[i][response.name] for result in all_results])

        best.append(entry)

    # then compute the regret from the best value
    quantiles = []

    for response in model.responses:
        if response.dir == Response.MINIMIZE or response.dir == Response.MAXIMIZE:
            values = []

            for i in range(len(all_results[0])):
                values.append(abs((baseline_results[i][response.name] - best[i][response.name]) / baseline_results[i][response.name]))

            quantiles.append(np.percentile(values, percentile))

    return max(quantiles)

def satisficing_type1(model, results, expr=None):
    if expr is None:
        return mean(check_feasibility(model, results))
    else:
        return [expr(result) for result in results]

def satisficing_type2(model, results, baseline, expr=None):
    distances = []

    # ensure all default parameters are defined in baseline
    baseline = baseline.copy()
    populate_defaults(model, baseline)

    for result in results:
        if (expr is None and _is_feasible(model, result)) or (expr is not None and expr(result)):
            distances.append(sp.distance.euclidean(
                    [result[u.name] for u in model.uncertainties],
                    [baseline[u.name] for u in model.uncertainties]))

    return 0.0 if len(distances) == 0 else min(distances)

def evaluate_robustness(model, policies, SOWs=1000, in_place=True, return_all=False):
    if isinstance(SOWs, int):
        SOWs = sample_lhs(model, SOWs)

    # evaluate the policies
    n = len(policies)
    all_results = [evaluate(model, update(SOWs, policy)) for policy in policies]

    # compute the standard robustness metrics
    result = policies if in_place else DataSet()

    result["Regret Type 1"] = [regret_type1(model, all_results[i], policies[i]) for i in range(n)]
    result["Regret Type 2"] = [regret_type2(model, all_results, all_results[i]) for i in range(n)]
    result["Satisficing Type 1"] = [satisficing_type1(model, all_results[i]) for i in range(n)]
    result["Satisficing Type 2"] = [satisficing_type2(model, all_results[i], policies[i]) for i in range(len(policies))]

    if return_all:
        return SOWs, result, all_results
    else:
        return result
